#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jul 10 16:04:15 2019

@author: annaquinlan

Github URL: https://github.com/tidepool-org/Loop/blob/
8c1dfdba38fbf6588b07cee995a8b28fcf80ef69/Loop/Managers/LoopDataManager.swift
"""
# pylint: disable=R0913, R0914, W0105, C0200, R0916
from datetime import timedelta
import warnings

from pyloopkit.carb_store import get_carb_glucose_effects, get_carbs_on_board
from pyloopkit.date import time_interval_since
from pyloopkit.dose import DoseType
from pyloopkit.dose_math import recommended_temp_basal, recommended_bolus
from pyloopkit.dose_store import get_glucose_effects
from pyloopkit.glucose_store import (get_recent_momentum_effects,
                           get_counteraction_effects)
from pyloopkit.input_validation_tools import (
    are_settings_valid, are_glucose_readings_valid, are_carb_readings_valid,
    is_insulin_sensitivity_schedule_valid, are_carb_ratios_valid,
    are_basal_rates_valid, are_correction_ranges_valid,
    are_insulin_doses_valid)
from pyloopkit.insulin_math import find_ratio_at_time
from pyloopkit.loop_math import (combined_sums, decay_effect, subtracting,
                       predict_glucose)


def update(input_dict):
    """ Run data through the Loop algorithm and return the predicted glucose
        values, recommended temporary basal, and recommended bolus

    Arguments:
    input_dict - dictionary containing the following keys/data:
        "glucose_dates" -- times of glucose measurements
        "glucose_values" -- glucose measurements in mg/dL)

        "dose_types" -- types of dose (tempBasal, bolus, etc)
        "dose_start_times" -- start times of insulin delivery
        "dose_end_times" -- end times of insulin delivery
        "dose_values" -- amounts of insulin (U/hr if a basal, U if a bolus)
        "dose_delivered_units" -- absolute Units of insulin delivered by a dose (can be None)

        "carb_dates" -- times of carbohydrate entries
        "carb_values" -- amount of carbohydrates eaten
        "carb_absorption_times" -- absorption times for carbohydrate entries)

        "settings_dictionary" -- a dictionary containing the needed settings:
            - "model" (the insulin model)
                - if exponential, format is
                    [duration of insulin action (in mins), peak (in mins)]
                        - child model typically peaks at 65 mins
                        - adult model typically peaks at 65 mins
                - if Walsh curve, format is
                    [duration of insulin action (in *hours*)]
            - "momentum_data_interval"
                - the interval of glucose data to use for momentum calculation
                - in Loop, default is 15 (minutes)
            - "suspend_threshold"
                - value at which to suspend all insulin delivery (mg/dL)
            - "dynamic_carb_absorption_enabled"
                - whether carb absorption can be calculated dynamically
                  (based on deviations in blood glucose levels versus what
                   would be expected based on insulin alone)
            - "retrospective_correction_integration_interval"
                - the maximum duration over which to integrate retrospective
                  correction changes
                - in Loop, default is 30 (minutes)
            - "recency_interval"
                - the amount of time since a given date that data should be
                  considered valid
                - in Loop, default is 15 (minutes)
            - "retrospective_correction_grouping_interval"
                - the interval over which to aggregate changes in glucose for
                  retrospective correction
                - in Loop, default is 30 (minutes)
            - "default_absorption_times"
                - the default absorption times to use if unspecified in a
                  carbohydrate entry
                - list of default absorption times in minutes in the format
                  [slow, medium, fast]
                - in Loop, default is [120 (fast), 180 (medium), 240 (slow)]
            - "max_basal_rate"
                - the maximum basal rate that Loop is allowed to give
            - "max_bolus"
                - the maximum bolus that Loop is allowed to give or recommend

        "sensitivity_ratio_start_times" -- start times for sensitivity ratios
        "sensitivity_ratio_end_times" -- end times for sensitivity ratios
        "sensitivity_ratio_values" -- the sensitivity ratios in mg/dL per U

        "carb_ratio_start_times" -- start times for carb ratios
        "carb_ratio_values" -- carb ratios in grams of carbohydrate per U

        "basal_rate_start_times" -- start times for basal rates
        "basal_rate_values" -- basal rates in U/hour
        "basal_rate_minutes" -- basal rate start minutes of offset from midnight

        "target_range_start_times" -- start times for the target ranges
        "target_range_end_times" -- end times for the target ranges
        "target_range_minimum_values" - minimum values of target range in mg/dL
        "target_range_maximum_values" - maximum values of target range in mg/dL

        last_temp_basal -- list of information about the last temporary basal
             in the form
             [type of dose,
              start time for the basal,
              end time for the basal,
              value of the basal rate in U/hr]

        time_to_calculate_at -- the "now" time and the time at which to
            recommend the basal rate and bolus

    Output:
        Dictionary containing all of the calculated effects, the input
        dictionary, the predicted glucose values, and the recommended
        temp basal/bolus
    """
    glucose_dates = input_dict.get("glucose_dates")
    glucose_values = input_dict.get("glucose_values")

    dose_types = input_dict.get("dose_types")
    dose_starts = input_dict.get("dose_start_times")
    dose_ends = input_dict.get("dose_end_times")
    dose_values = input_dict.get("dose_values")
    dose_delivered_units = input_dict.get("dose_delivered_units")

    carb_dates = input_dict.get("carb_dates")
    carb_values = input_dict.get("carb_values")
    carb_absorptions = input_dict.get("carb_absorption_times")

    settings_dictionary = input_dict.get("settings_dictionary")

    sensitivity_starts = input_dict.get("sensitivity_ratio_start_times")
    sensitivity_ends = input_dict.get("sensitivity_ratio_end_times")
    sensitivity_values = input_dict.get("sensitivity_ratio_values")

    carb_ratio_starts = input_dict.get("carb_ratio_start_times")
    carb_ratio_values = input_dict.get("carb_ratio_values")

    basal_starts = input_dict.get("basal_rate_start_times")
    basal_rates = input_dict.get("basal_rate_values")
    basal_minutes = input_dict.get("basal_rate_minutes")

    target_range_starts = input_dict.get("target_range_start_times")
    target_range_ends = input_dict.get("target_range_end_times")
    target_range_mins = input_dict.get("target_range_minimum_values") or []
    target_range_maxes = input_dict.get("target_range_maximum_values")

    last_temp_basal = input_dict.get("last_temporary_basal")

    time_to_calculate_at = input_dict.get("time_to_calculate_at")

    # check that the inputs make sense before doing math with them
    if (
            not are_settings_valid(settings_dictionary)
            or not are_glucose_readings_valid(
                glucose_dates, glucose_values,
            )
            or not are_carb_readings_valid(
                carb_dates, carb_values, carb_absorptions
            )

            or not are_insulin_doses_valid(
                dose_types, dose_starts, dose_ends, dose_values
            )
            or not is_insulin_sensitivity_schedule_valid(
                sensitivity_starts, sensitivity_ends, sensitivity_values
            )
            or not are_carb_ratios_valid(
                carb_ratio_starts, carb_ratio_values
            )
            or not are_basal_rates_valid(
                basal_starts, basal_rates, basal_minutes
            )
            or not are_correction_ranges_valid(
                target_range_starts, target_range_ends,
                target_range_mins, target_range_maxes
            )):
        return []

    last_glucose_date = glucose_dates[-1]

    retrospective_start = (
        last_glucose_date
        - timedelta(minutes=settings_dictionary.get(
            "retrospective_correction_integration_interval") or 30)
    )

    # calculate a maximum of 24 hours of effects
    earliest_effect_date = time_to_calculate_at - timedelta(hours=24)
    # if running with counteraction effect data present from a previous run
    # (which is how Loop runs), add the dates to the input dictionary
    next_effect_date = (
        input_dict.get("previous_counteraction_effect_dates")[-1] if
        input_dict.get("previous_counteraction_effect_dates")
        else earliest_effect_date
    )

    # Allow input effects to be passed through dict for testing purposes
    if (input_dict.get("momentum_effect_dates") and 
        input_dict.get("momentum_effect_values")):
        momentum_effect_dates = input_dict.get("momentum_effect_dates")
        momentum_effect_values = input_dict.get("momentum_effect_values")
    else:
        (momentum_effect_dates,
        momentum_effect_values
        ) = get_recent_momentum_effects(
            glucose_dates, glucose_values,
            next_effect_date,
            time_to_calculate_at,
            settings_dictionary.get("momentum_data_interval") or 15,
            5
            )

    # calculate previous insulin effects in order to later calculate the
    # insulin counteraction effects
    (insulin_effect_dates,
     insulin_effect_values
     ) = get_glucose_effects(
         dose_types, dose_starts, dose_ends, dose_values, dose_delivered_units,
         next_effect_date,
         basal_starts, basal_rates, basal_minutes,
         sensitivity_starts, sensitivity_ends, sensitivity_values,
         settings_dictionary.get("model"),
         delay=settings_dictionary.get("insulin_delay") or 10
         )

    # calculate future insulin effects for the purposes of predicting glucose
    if (input_dict.get("now_to_dia_insulin_effect_dates") and 
        input_dict.get("now_to_dia_insulin_effect_values")):
        now_to_dia_insulin_effect_dates = input_dict.get("now_to_dia_insulin_effect_dates")
        now_to_dia_insulin_effect_values = input_dict.get("now_to_dia_insulin_effect_values")
    else:
        (now_to_dia_insulin_effect_dates,
        now_to_dia_insulin_effect_values
        ) = get_glucose_effects(
            dose_types, dose_starts, dose_ends, dose_values, dose_delivered_units,
            time_to_calculate_at,
            basal_starts, basal_rates, basal_minutes,
            sensitivity_starts, sensitivity_ends, sensitivity_values,
            settings_dictionary.get("model"),
            delay=settings_dictionary.get("insulin_delay") or 10
            )

    
    if (input_dict.get("counteraction_starts") and 
        input_dict.get("counteraction_ends") and
        input_dict.get("counteraction_values")):
        counteraction_starts = input_dict.get("counteraction_starts")
        counteraction_ends = input_dict.get("counteraction_ends")
        counteraction_values = input_dict.get("counteraction_values")
        counteraction_effects = (counteraction_starts, counteraction_ends, counteraction_values)

    # if our BG data is current and we know the expected insulin effects,
    # calculate tbe counteraction effects
    elif next_effect_date < last_glucose_date and insulin_effect_dates:
        (counteraction_starts,
         counteraction_ends,
         counteraction_values
         ) = counteraction_effects = get_counteraction_effects(
             glucose_dates, glucose_values,
             next_effect_date,
             insulin_effect_dates, insulin_effect_values
             )
    else:
        (counteraction_starts,
         counteraction_ends,
         counteraction_values
         ) = counteraction_effects = ([], [], [])

    if (input_dict.get("carb_effect_dates") and 
        input_dict.get("carb_effect_values")):
        carb_effect_dates = input_dict.get("carb_effect_dates")
        carb_effect_values = input_dict.get("carb_effect_values")
    else: 
        (carb_effect_dates,
        carb_effect_values
        ) = get_carb_glucose_effects(
            carb_dates, carb_values, carb_absorptions,
            retrospective_start,
            *counteraction_effects if
            settings_dictionary.get("dynamic_carb_absorption_enabled")
            is not False else ([], [], []),
            carb_ratio_starts, carb_ratio_values,
            sensitivity_starts, sensitivity_ends, sensitivity_values,
            settings_dictionary.get("default_absorption_times"),
            delay=settings_dictionary.get("carb_delay") or 10
            )

    (cob_dates,
     cob_values
     ) = get_carbs_on_board(
         carb_dates, carb_values, carb_absorptions,
         time_to_calculate_at,
         *counteraction_effects if
         settings_dictionary.get("dynamic_carb_absorption_enabled")
         is not False else ([], [], []),
         carb_ratio_starts, carb_ratio_values,
         sensitivity_starts, sensitivity_ends, sensitivity_values,
         settings_dictionary.get("default_absorption_times"),
         delay=settings_dictionary.get("carb_delay") or 10
         )

    current_cob = cob_values[
        closest_prior_to_date(
            time_to_calculate_at,
            cob_dates
            )
        ] if cob_dates else 0

    if settings_dictionary.get("retrospective_correction_enabled"):
        (retrospective_effect_dates,
         retrospective_effect_values
         ) = update_retrospective_glucose_effect(
            glucose_dates, glucose_values,
            carb_effect_dates, carb_effect_values,
            counteraction_starts, counteraction_ends, counteraction_values,
            settings_dictionary.get("recency_interval") or 15,
            settings_dictionary.get(
                "retrospective_correction_grouping_interval"
            ) or 30,
            time_to_calculate_at
            )
    else:
        (retrospective_effect_dates,
         retrospective_effect_values
         ) = ([], [])

    recommendations = update_predicted_glucose_and_recommended_basal_and_bolus(
        time_to_calculate_at,
        glucose_dates, glucose_values,
        momentum_effect_dates, momentum_effect_values,
        carb_effect_dates, carb_effect_values,
        now_to_dia_insulin_effect_dates, now_to_dia_insulin_effect_values,
        retrospective_effect_dates, retrospective_effect_values,
        target_range_starts, target_range_ends, target_range_mins, target_range_maxes,
        settings_dictionary.get("suspend_threshold"),
        sensitivity_starts, sensitivity_ends, sensitivity_values,
        settings_dictionary.get("model"),
        basal_starts, basal_rates, basal_minutes,
        settings_dictionary.get("max_basal_rate"),
        settings_dictionary.get("max_bolus"),
        last_temp_basal,
        rate_rounder=settings_dictionary.get("rate_rounder")
        )

    recommendations["insulin_effect_dates"] = now_to_dia_insulin_effect_dates
    recommendations["insulin_effect_values"] = now_to_dia_insulin_effect_values

    recommendations["counteraction_effect_start_times"] = counteraction_starts
    recommendations["counteraction_effect_end_times"] = counteraction_ends
    recommendations["counteraction_effect_values"] = counteraction_values

    recommendations["momentum_effect_dates"] = momentum_effect_dates
    recommendations["momentum_effect_values"] = momentum_effect_values

    recommendations["carb_effect_dates"] = carb_effect_dates
    recommendations["carb_effect_values"] = carb_effect_values

    recommendations["retrospective_effect_dates"] = (
        retrospective_effect_dates or None
    )
    recommendations["retrospective_effect_values"] = (
        retrospective_effect_values or None
    )
    recommendations["carbs_on_board"] = current_cob
    recommendations["cob_timeline_dates"] = cob_dates
    recommendations["cob_timeline_values"] = cob_values
    recommendations["input_data"] = input_dict

    return recommendations


def closest_prior_to_date(date_to_compare, dates):
    """ Returns the index of the closest element in the sorted sequence
        prior to the specified date
    """
    for date in dates:
        if date <= date_to_compare:
            closest_element = date
        else:
            break

    return dates.index(closest_element)


def update_retrospective_glucose_effect(
        glucose_dates, glucose_values,
        carb_effect_dates, carb_effect_values,
        counteraction_starts, counteraction_ends, counteraction_values,
        recency_interval,
        retrospective_correction_grouping_interval,
        now_time,
        effect_duration=60,
        delta=5
        ):
    """
    Generate an effect based on how large the discrepancy is between the
    current glucose and its predicted value.

    Arguments:
    glucose_dates -- time of glucose value (datetime)
    glucose_values -- value at the time of glucose_date

    carb_effect_dates -- date the carb effects occur at (datetime)
    carb_effect_values -- value of carb effect

    counteraction_starts -- start times for counteraction effects
    counteraction_ends -- end times for counteraction effects
    counteraction_values -- values of counteraction effects

    recency_interval -- amount of time since a given date that data should be
                        considered valid
    retrospective_correction_grouping_interval -- interval over which to
        aggregate changes in glucose for retrospective correction

    now_time -- the time the loop is being run at
    effect_duration -- the length of time to calculate the retrospective
                       glucose effect out to
    delta -- time interval between glucose values (mins)

    Output:
    Retrospective glucose effect information in format
    (retrospective_effect_dates, retrospective_effect_values)
    """
    assert len(glucose_dates) == len(glucose_values),\
        "expected input shapes to match"

    assert len(carb_effect_dates) == len(carb_effect_values),\
        "expected input shapes to match"

    assert len(counteraction_starts) == len(counteraction_ends)\
        == len(counteraction_values), "expected input shapes to match"

    if not glucose_dates or not carb_effect_dates or not counteraction_starts:
        return ([], [])

    (discrepancy_starts, discrepancy_values) = subtracting(
        counteraction_starts, counteraction_ends, counteraction_values,
        carb_effect_dates, [], carb_effect_values,
        delta
        )

    retrospective_glucose_discrepancies_summed = combined_sums(
                discrepancy_starts, discrepancy_starts, discrepancy_values,
                retrospective_correction_grouping_interval * 1.01
                )

    # Our last change should be recent, otherwise clear the effects
    if (time_interval_since(
            now_time,
            retrospective_glucose_discrepancies_summed[1][-1])
            > recency_interval * 60
       ):
        return ([], [])

    discrepancy_time = max(
        0,
        retrospective_correction_grouping_interval
        )

    velocity = (
        retrospective_glucose_discrepancies_summed[2][-1]
        / discrepancy_time
        )

    return decay_effect(
        glucose_dates[-1], glucose_values[-1],
        velocity,
        effect_duration
        )


def get_pending_insulin(
        at_date,
        basal_starts, basal_rates, basal_minutes,
        last_temp_basal,
        pending_bolus_amount=None
    ):
    """ Get the pending insulin for the purposes of calculating a recommended
        bolus

    Arguments:
    at_date -- the "now" time (roughly equivalent to datetime.now)

    basal_starts -- list of times the basal rates start at
    basal_rates -- list of basal rates (U/hr)
    basal_minutes -- list of basal lengths (in mins)

    last_temp_basal -- information about the last temporary basal in the form
                       [type, start time, end time, basal rate]
    pending_bolus_amount -- amount of unconfirmed bolus insulin (U)

    Output:
    Amount of insulin that is "pending"
    """
    assert len(basal_starts) == len(basal_rates),\
        "expected input shapes to match"

    if (not basal_starts
            or not last_temp_basal
            or last_temp_basal[1] > last_temp_basal[2]
       ):
        return 0

    # if the end date for the temp basal is greater than current date,
    # find the pending insulin
    if (last_temp_basal[2] > at_date
            and last_temp_basal[0] in [DoseType.tempbasal, DoseType.basal]):
        normal_basal_rate = find_ratio_at_time(
            basal_starts, [], basal_rates, at_date
        )
        remaining_time = time_interval_since(
            last_temp_basal[2],
            at_date
        ) / 60 / 60

        remaining_units = (
            last_temp_basal[3] - normal_basal_rate
        ) * remaining_time
        pending_basal_insulin = max(0, remaining_units)

    else:
        pending_basal_insulin = 0

    if pending_bolus_amount:
        pending_bolus = pending_bolus_amount
    else:
        pending_bolus = 0

    return pending_basal_insulin + pending_bolus


def update_predicted_glucose_and_recommended_basal_and_bolus(
        at_date,
        glucose_dates, glucose_values,
        momentum_dates, momentum_values,
        carb_effect_dates, carb_effect_values,
        insulin_effect_dates, insulin_effect_values,
        retrospective_effect_dates, retrospective_effect_values,
        target_starts, target_ends, target_mins, target_maxes,
        suspend_threshold,
        sensitivity_starts, sensitivity_ends, sensitivity_values,
        model,
        basal_starts, basal_rates, basal_minutes,
        max_basal_rate, max_bolus,
        last_temp_basal,
        duration=30,
        continuation_interval=11,
        rate_rounder=None
        ):
    """ Generate glucose predictions, then use the predicted glucose along
        with settings and dose data to recommend a temporary basal rate and
        a bolus

    Arguments:
    at_date -- date to calculate the temp basal and bolus recommendations

    glucose_dates -- dates of glucose values (datetime)
    glucose_values -- glucose values (in mg/dL)

    momentum_dates -- times of calculated momentums (datetime)
    momentum_values -- values (mg/dL) of momentums

    carb_effect_dates -- times of carb effects (datetime)
    carb_effect -- values (mg/dL) of effects from carbs

    insulin_effect_dates -- times of insulin effects (datetime)
    insulin_effect -- values (mg/dL) of effects from insulin

    correction_effect_dates -- times of retrospective effects (datetime)
    correction_effect -- values (mg/dL) retrospective glucose effects

    target_starts -- start times for given target ranges (datetime)
    target_ends -- stop times for given target ranges (datetime)
    target_mins -- the lower bounds of target ranges (mg/dL)
    target_maxes -- the upper bounds of target ranges (mg/dL)

    suspend_threshold -- value at which to suspend all insulin delivery (mg/dL)

    sensitivity_starts -- list of time objects of start times of
                          given insulin sensitivity values
    sensitivity_ends -- list of time objects of start times of
                        given insulin sensitivity values
    sensitivity_values -- list of sensitivities (mg/dL/U)

    model -- list of insulin model parameters in format [DIA, peak_time] if
             exponential model, or [DIA] if Walsh model

    basal_starts -- list of times the basal rates start at
    basal_rates -- list of basal rates(U/hr)
    basal_minutes -- list of basal lengths (in mins)

    max_basal_rate -- max basal rate that Loop can give (U/hr)
    max_bolus -- max bolus that Loop can give (U)

    last_temp_basal -- list of last temporary basal information in format
                       [type, start time, end time, basal rate]
    duration -- length of the temp basal (mins)
    continuation_interval -- length of time before an ongoing temp basal
                             should be continued with a new command (mins)
    rate_rounder -- the smallest fraction of a unit supported in basal
                    delivery; if None, no rounding is performed

    Output:
    The predicted glucose values, recommended temporary basal, and
    recommended bolus in the format [
        (predicted glucose times, predicted glucose values),
        temporary basal recommendation,
        bolus recommendation
    ]
    """
    assert glucose_dates, "expected to receive glucose data"

    assert target_starts and sensitivity_starts and basal_starts and model,\
        "expected to receive complete settings data"

    if (not momentum_dates
            and not carb_effect_dates
            and not insulin_effect_dates
       ):
        warnings.warn("Warning: expected to receive effect data")
        return (None, None, None)

    predicted_glucoses = predict_glucose(
        glucose_dates[-1], glucose_values[-1],
        momentum_dates, momentum_values,
        carb_effect_dates, carb_effect_values,
        insulin_effect_dates, insulin_effect_values,
        retrospective_effect_dates, retrospective_effect_values
        )

    # Dosing requires prediction entries at least as long as the insulin
    # model duration. If our prediction is shorter than that, extend it here.
    if len(model) == 1:  # Walsh model
        final_date = glucose_dates[-1] + timedelta(hours=model[0])
    else:
        final_date = glucose_dates[-1] + timedelta(minutes=model[0])

    if predicted_glucoses[0][-1] < final_date:
        predicted_glucoses[0].append(final_date)
        predicted_glucoses[1].append(predicted_glucoses[1][-1])

    pending_insulin = get_pending_insulin(
        at_date,
        basal_starts, basal_rates, basal_minutes,
        last_temp_basal
    )

    temp_basal = recommended_temp_basal(
        *predicted_glucoses,
        target_starts, target_ends, target_mins, target_maxes,
        at_date,
        suspend_threshold,
        sensitivity_starts, sensitivity_ends, sensitivity_values,
        model,
        basal_starts, basal_rates, basal_minutes,
        max_basal_rate,
        last_temp_basal,
        duration,
        continuation_interval,
        rate_rounder
        )

    bolus = recommended_bolus(
        *predicted_glucoses,
        target_starts, target_ends, target_mins, target_maxes,
        at_date,
        suspend_threshold,
        sensitivity_starts, sensitivity_ends, sensitivity_values,
        model,
        pending_insulin,
        max_bolus,
        rate_rounder
        )

    return {
        "predicted_glucose_dates": predicted_glucoses[0],
        "predicted_glucose_values": predicted_glucoses[1],
        "recommended_temp_basal": temp_basal,
        "recommended_bolus": bolus
    }



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jun 14 08:45:26 2019

@author: annaquinlan

Github URL: https://github.com/tidepool-org/LoopKit/blob/
57a9f2ba65ae3765ef7baafe66b883e654e08391/LoopKit/Extensions/Date.swift
"""
import datetime
import math

REF_TIME = datetime.datetime.fromisoformat("2001-01-01T00:00:00")
TIMEZONE_REF_TIME = datetime.datetime.strptime(
    "2001-01-01 00:00:00 +0000",
    "%Y-%m-%d %H:%M:%S %z"
    )


def time_interval_since_reference_date(actual_time):
    """ Calculate seconds since since January, 1st, 2001 @ 12:00 AM

    Arguments:
    actual_time -- datetime object to compare to Jan 1st, 2001 @ 12:00 AM

    Output:
    Num of seconds since Jan 1st, 2001 @ 12:00 AM (without a sign)
    """
    if actual_time.tzinfo:
        return abs(actual_time - TIMEZONE_REF_TIME).total_seconds()

    return abs(actual_time - REF_TIME).total_seconds()


def time_interval_since(date_1, date_2):
    """ Calculate seconds between two times

    Arguments:
    date_1 -- datetime object #1
    date_2 -- datetime object #2

    Output:
    Num of seconds between the two times (with a sign)
    """
    return (date_1 - date_2).total_seconds()


def date_floored_to_time_interval(time_, interval):
    """ Floor a datetime object to a particular minute interval

    Arguments:
    time -- datetime object to be floored
    interval -- interval to floor the time to, measured in minutes

    Output:
    Floored datetime object

    Example:
        2/2/19 2:03 PM, interval=5 -> 2/2/19 2:00 PM
    """
    if interval == 0:
        return time_

    floored_delta = (math.floor(time_interval_since_reference_date(time_) /
                                interval / 60) * interval * 60)

    if time_.tzinfo:
        return TIMEZONE_REF_TIME + datetime.timedelta(seconds=floored_delta)

    return REF_TIME + datetime.timedelta(seconds=floored_delta)


def date_ceiled_to_time_interval(time, interval):
    """ Ceil a datetime object to a particular minute interval

    Arguments:
    time -- datetime object to be ceiled
    interval -- interval to ceil the time to, measured in minutes

    Output:
    Ceiled datetime object

    Example:
        2/2/19 2:03 PM, interval=5 -> 2/2/19 2:05 PM
    """
    if interval == 0:
        return time

    ceiled_delta = (math.ceil(time_interval_since_reference_date(time) /
                              interval / 60) * interval * 60)

    if time.tzinfo:
        return TIMEZONE_REF_TIME + datetime.timedelta(seconds=ceiled_delta)

    return REF_TIME + datetime.timedelta(seconds=ceiled_delta)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jul  9 18:04:24 2019

@author: annaquinlan

Github URL: https://github.com/tidepool-org/Loop/blob/
8c1dfdba38fbf6588b07cee995a8b28fcf80ef69/Loop/Managers/LoopDataManager.swift
"""
# pylint: disable=R0913, R0914, C0200, R1705, R0912
from datetime import timedelta
from enum import Enum
import sys

from pyloopkit.insulin_math import is_time_between, find_ratio_at_time
from pyloopkit.date import time_interval_since
from pyloopkit.dose import DoseType
from pyloopkit.walsh_insulin_model import walsh_percent_effect_remaining
from pyloopkit.exponential_insulin_model import percent_effect_remaining


class Correction(Enum):
    suspend = 0
    in_range = 1
    above_range = 2
    entirely_below_range = 3
    cancel = 4


def filter_date_range_for_doses(
        types, starts, ends, values, delivered_units,
        start_date,
        end_date
        ):
    """ Returns an array of elements filtered by the specified date range.

    Arguments:
    types -- String of type of dose (basal, bolus, etc)
    starts -- start dates (datetime)
    ends -- end dates (datetime)
    values -- glucose values

    start_date -- the earliest date of elements to return
    end_date -- the last date of elements to return

    Output:
    Filtered dates in format (types, starts, ends, values)
    """
    # ends might not necesarily be the same length as starts/values
    # because not all types have "end dates"
    assert len(types) == len(starts) == len(values) == len(delivered_units),\
        "expected input shapes to match"

    (filtered_types,
     filtered_starts,
     filtered_ends,
     filtered_values,
     filtered_delivered_units
     ) = ([], [], [], [], [])

    for i in range(0, len(starts)):
        if start_date and ends and ends[i] < start_date:
            continue

        if start_date and not ends and starts[i] < start_date:
            continue

        if end_date and starts[i] > end_date:
            continue

        filtered_types.append(types[i])
        filtered_starts.append(starts[i])
        filtered_ends.append(ends[i] if ends else None)
        filtered_values.append(values[i])
        filtered_delivered_units.append(delivered_units[i])

    assert len(filtered_types) == len(filtered_starts) == len(filtered_ends)\
        == len(filtered_values) == len(filtered_delivered_units),\
        "expected output shapes to match"

    return (
        filtered_types,
        filtered_starts,
        filtered_ends,
        filtered_values,
        filtered_delivered_units
        )


def target_glucose_value(
        percent_effect_duration,
        min_value,
        max_value
        ):
    """ Computes a target glucose value for a correction, at a given time
        during the insulin effect duration

    Arguments:
    percent_effect_duration -- percent of time elapsed of the insulin
                               effect duration
    min_value -- minimum (starting) target value
    max_value -- maximum (starting) target value

    Output:
    A target value somewhere between the minimum and maximum (inclusive)
    """
    # The inflection point in time: before it we use minValue,
    # after it we linearly blend from minValue to maxValue
    use_min_value_until_percent = 0.5

    if percent_effect_duration <= use_min_value_until_percent:
        return min_value

    if percent_effect_duration >= 1:
        return max_value

    slope = (
        (max_value - min_value) /
        (1 - use_min_value_until_percent)
        )

    return min_value + slope * (percent_effect_duration
                                - use_min_value_until_percent
                                )


def insulin_correction_units(
        from_value,
        to_value,
        effected_sensitivity
        ):
    """ Computes a total insulin amount necessary to correct a glucose
        differential at a given sensitivity

    Arguments:
    from_value -- starting glucose value
    to_value -- desired glucose value
    effected_sensitivity -- sensitivity (mg/dL/U)

    Output:
    The insulin correction in units
    """
    if effected_sensitivity <= 0:
        return None

    glucose_correction = from_value - to_value
    return glucose_correction / effected_sensitivity


def matches_rate(rate_1, rate_2):
    """ Determine if two rates are the same """
    return abs(rate_1 - rate_2) < sys.float_info.epsilon


def if_necessary(
        temp_basal,
        at_date,
        scheduled_basal_rate,
        last_temp_basal,
        continuation_interval
        ):
    """ Determine whether the recommendation is necessary given the
        current state of the pump

    Arguments:
    temp_basal -- recommended temp basal
    at_date -- date to calculate temp basal at (datetime)
    scheduled_basal_rate -- basal rate scheduled during "at_date"
    last_temp_basal -- the previously set temp basal
    continuation_interval -- duration of time before an ongoing temp basal
                             should be continued with a new command

    Output:
    None (if the scheduled temp basal or basal rate should be allowed to
    continue to run), cancel (if the temp basal should be cancelled),
    or the recommended temp basal (if it should be set)
    """
    # Adjust behavior for the currently active temp basal
    if (last_temp_basal
            and last_temp_basal[0] in [DoseType.tempbasal, DoseType.basal]
            and last_temp_basal[2] > at_date
       ):
        # If the last temp basal has the same rate, and has more than
        # "continuation_interval" of time remaining, don't set a new temp
        if (matches_rate(temp_basal[0], last_temp_basal[3])
                and (
                    (time_interval_since(last_temp_basal[2], at_date) / 60)
                    > continuation_interval)
           ):
            return None

        # If our new temp matches the scheduled rate, cancel the current temp
        elif matches_rate(temp_basal[0], scheduled_basal_rate):
            return Correction.cancel

    # If we recommend the in-progress scheduled basal rate, do nothing
    elif matches_rate(temp_basal[0], scheduled_basal_rate):
        return None

    return temp_basal


def insulin_correction(
        prediction_dates, prediction_values,
        target_starts, target_ends, target_mins, target_maxes,
        at_date,
        suspend_threshold_value,
        sensitivity_value,
        model
        ):
    """ Computes a total insulin amount necessary to correct a glucose
        differential at a given sensitivity

    prediction_dates -- dates glucose values were predicted (datetime)
    prediction_values -- predicted glucose values (mg/dL)

    target_starts -- start times for given target ranges (datetime)
    target_ends -- stop times for given target ranges (datetime)
    target_mins -- the lower bounds of target ranges (mg/dL)
    target_maxes -- the upper bounds of target ranges (mg/dL)

    at_date -- date to calculate correction
    suspend_threshold -- value to suspend all insulin delivery at (mg/dL)
    sensitivity_value -- the sensitivity (mg/dL/U)
    model -- list of insulin model parameters in format [DIA, peak_time] if
             exponential model, or [DIA] if Walsh model

    Output:
    A list of insulin correction information. All lists have the type as the
    first index, and may include additional information based on the type.

        Types:
        - entirely_below_range
            Structure: [type, glucose value to be corrected, minimum target,
                        units of correction insulin]
        - suspend
            Structure: [type, min glucose value]
        - in_range
            Structure: [type]
        - above_range
            Structure: [type, minimum predicted glucose value,
                        glucose value to be corrected, minimum target,
                        units of correction insulin]
    """
    assert len(prediction_dates) == len(prediction_values),\
        "expected input shapes to match"

    assert len(target_starts) == len(target_ends) == len(target_mins)\
        == len(target_maxes), "expected input shapes to match"

    (min_glucose,
     eventual_glucose,
     correcting_glucose,
     min_correction_units
     ) = ([], None, None, None)

    # only calculate a correction if the prediction is between
    # "now" and now + DIA
    if len(model) == 1:  # if Walsh model
        date_range = [at_date,
                      at_date + timedelta(hours=model[0])
                      ]
    else:
        date_range = [at_date,
                      at_date + timedelta(minutes=model[0])
                      ]

    # if we don't know the suspend threshold, it defaults to the lower
    # bound of the correction range at the time the "loop" is being run at
    if not suspend_threshold_value:
        suspend_threshold_value = find_ratio_at_time(
            target_starts,
            target_ends,
            target_mins,
            at_date
            )

    # For each prediction above target, determine the amount of insulin
    # necessary to correct glucose based on the modeled effectiveness of
    # the insulin at that time
    for i in range(0, len(prediction_dates)):
        if not is_time_between(
                date_range[0],
                date_range[1],
                prediction_dates[i]
            ):
            continue

        # If any predicted value is below the suspend threshold,
        # return immediately
        if prediction_values[i] < suspend_threshold_value:
            return [Correction.suspend, prediction_values[i]]

        # Update range statistics
        if not min_glucose or prediction_values[i] < min_glucose[1]:
            min_glucose = [prediction_dates[i], prediction_values[i]]

        eventual_glucose = [prediction_dates[i], prediction_values[i]]
        predicted_glucose_value = prediction_values[i]
        time = time_interval_since(
            prediction_dates[i],
            at_date
            ) / 60

        average_target = (
            find_ratio_at_time(
                target_starts,
                target_ends,
                target_maxes,
                prediction_dates[i]
                ) +
            find_ratio_at_time(
                target_starts,
                target_ends,
                target_mins,
                prediction_dates[i]
                )
            ) / 2
        # Compute the target value as a function of time since the dose started
        target_value = target_glucose_value(
            (time /
             (
                 (60 * model[0]) if len(model) == 1
                 else model[0]
             )
            ),
            suspend_threshold_value,
            average_target
        )

        # Compute the dose required to bring this prediction to target:
        # dose = (Glucose delta) / (% effect × sensitivity)
        if len(model) == 1:  # if Walsh model
            percent_effected = 1 - walsh_percent_effect_remaining(
                time,
                model[0]
            )
        else:
            percent_effected = 1 - percent_effect_remaining(
                time,
                model[0],
                model[1]
            )
        effected_sensitivity = percent_effected * sensitivity_value

        # calculate the Units needed to correct that predicted glucose value
        correction_units = insulin_correction_units(
            predicted_glucose_value,
            target_value,
            effected_sensitivity
        )

        if not correction_units or correction_units <= 0:
            continue

        # Update the correction only if we've found a new minimum
        if min_correction_units:
            if correction_units >= min_correction_units:
                continue

        correcting_glucose = [prediction_dates[i], prediction_values[i]]
        min_correction_units = correction_units

    if not eventual_glucose or not min_glucose:
        return None

    # Choose either the minimum glucose or eventual glucose as correction delta
    min_glucose_targets = [
        find_ratio_at_time(
            target_starts,
            target_ends,
            target_mins,
            min_glucose[0]
        ),
        find_ratio_at_time(
            target_starts,
            target_ends,
            target_maxes,
            min_glucose[0]
        )
    ]
    eventual_glucose_targets = [
        find_ratio_at_time(
            target_starts,
            target_ends,
            target_mins,
            eventual_glucose[0]
            ),
        find_ratio_at_time(
            target_starts,
            target_ends,
            target_maxes,
            eventual_glucose[0]
        )
    ]

    # Treat the mininum glucose when both are below range
    if (min_glucose[1] < min_glucose_targets[0]
            and eventual_glucose[1] < min_glucose_targets[0]
       ):
        time = time_interval_since(min_glucose[0], at_date) / 60
        # For time = 0, assume a small amount effected.
        # This will result in large (negative) unit recommendation
        # rather than no recommendation at all.
        if len(model) == 1:
            percent_effected = max(
                sys.float_info.epsilon,
                1 - walsh_percent_effect_remaining(time, model[0])
                )
        else:
            percent_effected = max(
                sys.float_info.epsilon,
                1 - percent_effect_remaining(time, model[0], model[1])
                )

        units = insulin_correction_units(
            min_glucose[1],
            sum(min_glucose_targets) / len(min_glucose_targets),
            sensitivity_value * percent_effected
            )

        if not units:
            return None

        # we're way below target
        return [
            Correction.entirely_below_range,
            min_glucose[1],
            min_glucose_targets[0],
            units
            ]

    # we're above target
    elif (eventual_glucose[1] > eventual_glucose_targets[1]
          and min_correction_units
          and correcting_glucose
         ):
        return [
            Correction.above_range,
            min_glucose[1],
            correcting_glucose[1],
            eventual_glucose_targets[0],
            min_correction_units
            ]
    # we're in range
    else:
        return [Correction.in_range]


def as_temp_basal(
        correction,
        scheduled_basal_rate,
        max_basal_rate,
        duration,
        rate_rounder=None
        ):
    """ Determine temp basal over duration needed to perform the correction

    Arguments:
    correction -- list of information about the total amount of insulin
                  needed to correct BGs
    scheduled_basal_rate -- scheduled basal rate at time correction is given
    max_basal_rate -- the maximum allowed basal rate
    duration -- duration of the temporary basal
    rate_rounder -- the smallest fraction of a unit supported in basal delivery

    Output:
    Temp basal recommendation in form [temp basal rate, temp basal duration]
    """
    rate = (
        correction[len(correction) - 1] / (duration / 60)
        if correction[0] in [
            Correction.entirely_below_range, Correction.above_range
        ] else 0
    )
    # if it's not a suspend, add in the scheduled basal rate
    if not correction[0] == Correction.suspend:
        rate += scheduled_basal_rate

    rate = min(max_basal_rate, max(0, rate))

    # round to the appropriate increment for the insulin pump
    if rate_rounder:
        rate_rounder = 1 / rate_rounder
        rate = round(rate * rate_rounder) / rate_rounder

    return [rate, duration]


def bolus_recommendation_notice(correction):
    """ Make a bolus recommendation based on an insulin correction """
    if correction[0] == Correction.suspend:
        return ["glucoseBelowSuspendThreshold", correction[1]]

    if correction[0] in [Correction.entirely_below_range, Correction.in_range]:
        return None

    if correction[0] == Correction.above_range:
        # if we're recommending units but the minimum glucose is below target
        if correction[4] > 0 and correction[1] < correction[3]:
            return ["predictedGlucoseBelowTarget", correction[1]]

    return None


def as_bolus(
        correction,
        pending_insulin,
        max_bolus,
        volume_rounder
        ):
    """  Determine the bolus needed to perform the correction

    Arguments:
    correction -- list of information about the total amount of insulin
                  needed to correct BGs
    pending_insulin -- number of units expected to be delivered, but not yet
                       reflected in the correction
    max_bolus -- the maximum allowed bolus
    volume_rounder -- the smallest fraction of a unit supported in
                      insulin delivery

    Output:
    A bolus recommendation in form [units of bolus, pending insulin,
                                    recommendation]
    """
    correction_units = (
        correction[len(correction) - 1]
        if correction[0] in [
            Correction.above_range, Correction.entirely_below_range
        ] else 0
    )
    units = correction_units - pending_insulin
    units = min(max_bolus, max(0, units))

    if volume_rounder:
        volume_rounder = 1 / volume_rounder
        units = round(units * volume_rounder) / volume_rounder

    recommendation = bolus_recommendation_notice(correction)

    return [units, pending_insulin, recommendation]


def recommended_temp_basal(
        glucose_dates, glucose_values,
        target_starts, target_ends, target_mins, target_maxes,
        at_date,
        suspend_threshold,
        sensitivity_starts, sensitivity_ends, sensitivity_values,
        model,
        basal_starts, basal_rates, basal_minutes,
        max_basal_rate,
        last_temp_basal,
        duration=30,
        continuation_interval=11,
        rate_rounder=None
        ):
    """ Recommends a temporary basal rate to conform a glucose prediction
    timeline to a correction range

    Returns None if normal scheduled basal or active temporary basal is
    sufficient

    Arguments:
    glucose_dates -- dates of glucose values (datetime)
    glucose_values -- glucose values (in mg/dL)

    target_starts -- start times for given target ranges (datetime)
    target_ends -- stop times for given target ranges (datetime)
    target_mins -- the lower bounds of target ranges (mg/dL)
    target_maxes -- the upper bounds of target ranges (mg/dL)

    at_date -- date to calculate the temp basal at
    suspend_threshold -- value to suspend all insulin delivery at (mg/dL)

    sensitivity_starts -- list of time objects of start times of
                          given insulin sensitivity values
    sensitivity_ends -- list of time objects of start times of
                        given insulin sensitivity values
    sensitivity_values -- list of sensitivities (mg/dL/U)

    model -- list of insulin model parameters in format [DIA, peak_time] if
             exponential model, or [DIA] if Walsh model

    basal_starts -- list of times the basal rates start at
    basal_rates -- list of basal rates(U/hr)
    basal_minutes -- list of basal lengths (in mins)

    max_basal_rate -- max basal rate that Loop can give (U/hr)
    last_temp_basal -- list of last temporary basal information in format
                       [type, start time, end time, basal rate]
    duration -- length of the temp basal (mins)
    continuation_interval -- length of time before an ongoing temp basal
                             should be continued with a new command (mins)
    rate_rounder -- the smallest fraction of a unit supported in basal
                    delivery; if None, no rounding is performed

    Output:
    The recommended temporary basal in the format [rate, duration]
    """
    assert len(glucose_dates) == len(glucose_values),\
        "expected input shapes to match"

    assert len(target_starts) == len(target_ends) == len(target_mins)\
        == len(target_maxes), "expected input shapes to match"

    assert len(sensitivity_starts) == len(sensitivity_ends)\
        == len(sensitivity_values), "expected input shapes to match"

    assert len(basal_starts) == len(basal_rates) == len(basal_minutes),\
        "expected input shapes to match"

    if (not glucose_dates
            or not target_starts
            or not sensitivity_starts
            or not basal_starts
       ):
        return None

    sensitivity_value = find_ratio_at_time(
        sensitivity_starts, sensitivity_ends, sensitivity_values,
        at_date
        )

    correction = insulin_correction(
        glucose_dates, glucose_values,
        target_starts, target_ends, target_mins, target_maxes,
        at_date,
        suspend_threshold,
        sensitivity_value,
        model
        )

    scheduled_basal_rate = find_ratio_at_time(
        basal_starts, [], basal_rates,
        at_date
        )

    if (correction[0] == Correction.above_range
            and correction[1] < correction[3]):
        max_basal_rate = scheduled_basal_rate

    temp_basal = as_temp_basal(
        correction,
        scheduled_basal_rate,
        max_basal_rate,
        duration,
        rate_rounder
        )

    recommendation = if_necessary(
        temp_basal,
        at_date,
        scheduled_basal_rate,
        last_temp_basal,
        continuation_interval
        )

    # convert a "cancel" into zero-temp, zero-duration basal
    if recommendation == Correction.cancel:
        return [0, 0]

    return recommendation


def recommended_bolus(
        glucose_dates, glucose_values,
        target_starts, target_ends, target_mins, target_maxes,
        at_date,
        suspend_threshold,
        sensitivity_starts, sensitivity_ends, sensitivity_values,
        model,
        pending_insulin,
        max_bolus,
        volume_rounder=None
        ):
    """ Recommends a temporary basal rate to conform a glucose prediction
    timeline to a correction range

    Returns None if normal scheduled basal or active temporary basal is
    sufficient

    Arguments:
    glucose_dates -- dates of glucose values (datetime)
    glucose_values -- glucose values (in mg/dL)

    target_starts -- start times for given target ranges (datetime)
    target_ends -- stop times for given target ranges (datetime)
    target_mins -- the lower bounds of target ranges (mg/dL)
    target_maxes -- the upper bounds of target ranges (mg/dL)

    at_date -- date to calculate the temp basal at
    suspend_threshold -- value to suspend all insulin delivery at (mg/dL)

    sensitivity_starts -- list of time objects of start times of
                          given insulin sensitivity values
    sensitivity_ends -- list of time objects of start times of
                        given insulin sensitivity values
    sensitivity_values -- list of sensitivities (mg/dL/U)

    model -- list of insulin model parameters in format [DIA, peak_time] if
             exponential model, or [DIA] if Walsh model

    pending_insulin -- number of units expected to be delivered, but not yet
                       reflected in the correction
    max_bolus -- the maximum allowable bolus value in Units
    volume_rounder -- the smallest fraction of a unit supported in insulin
                      delivery; if None, no rounding is performed

    Output:
    A bolus recommendation
    """
    assert len(glucose_dates) == len(glucose_values),\
        "expected input shapes to match"

    assert len(target_starts) == len(target_ends) == len(target_mins)\
        == len(target_maxes), "expected input shapes to match"

    assert len(sensitivity_starts) == len(sensitivity_ends)\
        == len(sensitivity_values), "expected input shapes to match"

    if (not glucose_dates
            or not target_starts
            or not sensitivity_starts
       ):
        return [0, 0, None]

    sensitivity_value = find_ratio_at_time(
        sensitivity_starts, sensitivity_ends, sensitivity_values,
        at_date
        )

    correction = insulin_correction(
        glucose_dates, glucose_values,
        target_starts, target_ends, target_mins, target_maxes,
        at_date,
        suspend_threshold,
        sensitivity_value,
        model
        )

    bolus = as_bolus(
        correction,
        pending_insulin,
        max_bolus,
        volume_rounder
        )

    if bolus[0] < 0:
        bolus = 0

    return bolus



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jun 20 10:29:59 2019

@author: annaquinlan

Github URL: https://github.com/tidepool-org/LoopKit/blob/
57a9f2ba65ae3765ef7baafe66b883e654e08391/LoopKit/InsulinKit/InsulinMath.swift
"""
# pylint: disable=R0913, R0914, R0912, C0200, R0915, R1702, C0302, R0911
from math import floor
from datetime import timedelta, datetime
import sys

from pyloopkit.date import time_interval_since, time_interval_since_reference_date
from pyloopkit.dose import DoseType
from pyloopkit.loop_math import simulation_date_range_for_samples
from pyloopkit.dose_entry import net_basal_units, total_units_given
from pyloopkit.exponential_insulin_model import percent_effect_remaining
from pyloopkit.walsh_insulin_model import walsh_percent_effect_remaining

MAXIMUM_RESERVOIR_DROP_PER_MINUTE = 6.5
DISTANT_PAST = datetime.fromisoformat("2001-01-01T00:00:00")
TIMEZONE_DISTANT_PAST = datetime.strptime(
    "2001-01-01 00:00:00 +0000",
    "%Y-%m-%d %H:%M:%S %z"
    )
DISTANT_FUTURE = datetime.fromisoformat("2050-01-01T00:00:00")
TIMEZONE_DISTANT_FUTURE = datetime.strptime(
    "2050-01-01 00:00:00 +0000",
    "%Y-%m-%d %H:%M:%S %z"
    )


def total_delivery(dose_types, starts, ends, values):
    """ Calculates the total insulin delivery for a collection of doses

    Arguments:
    dose_types -- types of doses (basal, bolus, etc)
    starts -- datetime objects of times doses started at
    ends -- datetime objects of times doses ended at
    values -- amount, in U/hr (if a basal) or U (if bolus) of insulin in dose

    Output:
    The total insulin insulin, in Units
    """
    assert len(dose_types) == len(starts) == len(ends) == len(values),\
        "expected input shapes to match"

    total = 0
    for i in range(0, len(dose_types)):
        total += total_units_given(
            dose_types[i],
            values[i],
            starts[i],
            ends[i]
        )

    if total < 0:
        return 0

    return total


def dose_entries(reservoir_dates, unit_volumes):
    """ Converts a continuous, chronological sequence of reservoir values
        to a sequence of doses
    Runtime: O(n)

    Arguments:
    reservoir_dates -- list of datetime objects
    unit_volumes -- list of reservoir volumes (in units of insulin)

    Output:
    A tuple of lists in (dose_type (basal/bolus), start_dates, end_dates,
        insulin_values) format
    """
    assert len(reservoir_dates) > 1,\
        "expected input lists to contain two or more items"
    assert len(reservoir_dates) == len(unit_volumes),\
        "expected input shape to match"

    dose_types = []
    start_dates = []
    end_dates = []
    insulin_values = []

    previous_date = reservoir_dates[0]
    previous_unit_volume = unit_volumes[0]

    for i in range(1, len(reservoir_dates)):
        volume_drop = previous_unit_volume - unit_volumes[i]
        duration = time_interval_since(
            reservoir_dates[i],
            previous_date
        )

        if (duration > 0 and 0 <= volume_drop <=
                MAXIMUM_RESERVOIR_DROP_PER_MINUTE * duration / 60):
            dose_types.append(DoseType.tempbasal)
            start_dates.append(previous_date)
            end_dates.append(reservoir_dates[i])
            insulin_values.append(volume_drop)

        previous_date = reservoir_dates[i]
        previous_unit_volume = unit_volumes[i]

    assert len(dose_types) == len(start_dates) == len(end_dates) ==\
        len(insulin_values), "expected output shape to match"

    return (dose_types, start_dates, end_dates, insulin_values)


def is_continuous(reservoir_dates, unit_volumes, start, end,
                  maximum_duration):
    """ Whether a span of chronological reservoir values is considered
        continuous and therefore reliable.

    Reservoir values of 0 are automatically considered unreliable due to
    the assumption that an unknown amount of insulin can be delivered after
    the 0 marker.

    Arguments:
    reservoir_dates -- list of datetime objects that correspond by index to
                        unit_volumes
    unit_volumes -- volume of reservoir in units, corresponds by index to
                    reservoir_dates
    start -- datetime object that is start of the interval which to validate
             continuity
    end -- datetime object that is end of the interval which to validate
             continuity
    maximum_duration -- the maximum interval to consider reliable for a
                        reservoir-derived dose

    Variable names:
    start_date -- the beginning of the interval in which to validate
                   continuity
    end_date -- the end of the interval in which to validate continuity

    Outputs:
    Whether the reservoir values meet the critera for continuity
    """
    try:
        first_date_value = reservoir_dates[0]
        first_volume_value = unit_volumes[0]
    except IndexError:
        return False

    if end < start:
        return False

    start_date = start
    # The first value has to be at least as old as the start date
    # as a reference point.
    if first_date_value > start_date:
        return False

    last_date_value = first_date_value
    last_volume_value = first_volume_value

    for i in range(0, len(unit_volumes)):  # pylint: disable=C0200
        # Volume and interval validation only applies for values in
        # the specified range
        if reservoir_dates[i] < start_date or reservoir_dates[i] > end:
            last_date_value = reservoir_dates[i]
            last_volume_value = unit_volumes[i]
            continue
        # We can't trust 0. What else was delivered?
        if unit_volumes[i] <= 0:
            return False
        # Rises in reservoir volume indicate a rewind + prime, and primes
        # can be easily confused with boluses.
        # Small rises (1 U) can be ignored as they're indicative of a
        # mixed-precision sequence.
        if unit_volumes[i] > last_volume_value + 1:
            return False
        # Ensure no more than the maximum interval has passed
        if (time_interval_since(reservoir_dates[i], last_date_value)
                > maximum_duration * 60):
            return False

        last_date_value = reservoir_dates[i]
        last_volume_value = unit_volumes[i]

    return True


def reconciled(dose_types, start_dates, end_dates, values, delivered_units):
    """ Maps a timeline of dose entries with overlapping start and end dates
        to a timeline of doses that represents actual insulin delivery.

    Arguments:
    dose_types -- list of types of doses (basal, bolus, etc)
    start_dates -- list of datetime objects representing the dates
                   the doses started at
    end_dates -- list of datetime objects representing the dates
                   the doses ended at
    values -- list of insulin values for doses
    scheduled_basal_rates -- basal rates scheduled during the times of doses

    Output:
    Tuple with *four* of the dose properties (does not include scheduled basal
    rates), reconciled as TempBasal and Bolus records; these lists are
    *not* always in order of time
    """
    # this function does not return a list of scheduled basal rates
    output_types = []
    output_starts = []
    output_ends = []
    output_values = []
    output_delivered_units = []

    last_suspend_index = None
    last_basal = []

    assert len(dose_types) == len(start_dates) == len(end_dates) ==\
        len(values) == len(delivered_units),\
        "expected input shapes to match"

    for (i, type_) in enumerate(dose_types):
        if type_ == DoseType.bolus:
            output_types.append(type_)
            output_starts.append(start_dates[i])
            output_ends.append(end_dates[i])
            output_values.append(values[i])
            output_delivered_units.append(delivered_units[i])

        elif type_ in [DoseType.tempbasal, DoseType.basal]:
            if last_basal and not last_suspend_index:
                last = last_basal
                end_date = min(last[2], start_dates[i])

                # ignore zero-duration doses
                if end_date > last[1]:
                    output_types.append(last[0])
                    output_starts.append(last[1])
                    output_ends.append(end_date)
                    output_values.append(last[3])
                    output_delivered_units.append(last[4])
            last_basal = [type_, start_dates[i], end_dates[i],
                          values[i], delivered_units[i]]

        elif type_ == DoseType.resume:
            if last_suspend_index:
                suspend = last_suspend_index

                output_types.append(dose_types[suspend])
                output_starts.append(start_dates[suspend])
                output_ends.append(end_dates[i])
                output_values.append(values[suspend])
                output_delivered_units.append(delivered_units[suspend])

                last_suspend_index = None

                # Continue temp basals that may have started before suspending
                if last_basal:
                    last = last_basal
                    if last[2] > end_dates[i]:
                        last_basal = [last[0], end_dates[i], last[2], last[3], last[4]]
                    else:
                        last_basal = []

        elif type_ == DoseType.suspend:
            if last_basal:
                last = last_basal

                output_types.append(last[0])
                output_starts.append(last[1])
                output_ends.append(min(last[2], start_dates[i]))
                output_values.append(last[3])
                output_delivered_units.append(last[4])

                if last[2] <= start_dates[i]:
                    last_basal = []

            # add the suspend immediately if it's already been normalized
            # before being passed into reconciled()
            if start_dates[i] == end_dates[i]:
                last_suspend_index = i
            else:
                output_types.append(type_)
                output_starts.append(start_dates[i])
                output_ends.append(end_dates[i])
                output_values.append(values[i])
                output_delivered_units.append(delivered_units[i])

        elif type_ == DoseType.meal:
            output_types.append(type_)
            output_starts.append(start_dates[i])
            output_ends.append(end_dates[i])
            output_values.append(values[i])
            output_delivered_units.append(delivered_units[i])

    if last_suspend_index is not None:
        output_types.append(dose_types[last_suspend_index])
        output_starts.append(start_dates[last_suspend_index])
        output_ends.append(end_dates[last_suspend_index])
        output_values.append(values[last_suspend_index])
        output_delivered_units.append(delivered_units[last_suspend_index])

    elif (last_basal
          and last_basal[2] > last_basal[1]
          ):
        # I slightly modified this because it wasn't dealing with the last
        # basal correctly
        output_types.append(last_basal[0])
        output_starts.append(last_basal[1])
        output_ends.append(last_basal[2])
        output_values.append(last_basal[3])
        output_delivered_units.append(last_basal[4])

    assert len(output_types) == len(output_starts) == len(output_ends) ==\
        len(output_values) == len(output_delivered_units), "expected output shape to match"

    return (output_types,
            output_starts,
            output_ends,
            output_values,
            output_delivered_units
            )


def annotated(
        dose_types, start_dates, end_dates, values, delivered_units,
        basal_start_times, basal_rates, basal_minutes,
        convert_to_units_hr=True
    ):
    """ Annotates doses with the context of the scheduled basal rates

    Arguments:
    dose_types -- list of types of dose (basal, bolus, etc)
    dose_start_dates -- start dates of the doses (datetime obj)
    dose_end_dates -- end dates of the doses (datetime obj)
    values -- actual basal rates of doses in U/hr (if a basal)
             or the value of the boluses in U

    basal_start_times -- list of times the basal rates start at
    basal_rates -- list of basal rates(U/hr)
    basal_minutes -- list of basal lengths (in mins)

    convert_to_units_hr -- set to True if you want to convert the doses to U/hr
        (ex: 0.05 U given from 1/1/01 1:00:00 to 1/1/01 1:05:00 -> 0.6 U/hr);
        this will normally be for reservoir values

    Output:
    6 lists of annotated dose properties
    """
    assert len(dose_types) == len(start_dates) == len(end_dates) ==\
        len(values) == len(delivered_units),\
        "expected input shapes to match"

    assert len(basal_start_times) == len(basal_rates) == len(basal_minutes),\
        "expected input shapes to match"

    if not dose_types or not basal_start_times:
        return ([], [], [], [], [], [])

    output_types = []
    output_start_dates = []
    output_end_dates = []
    output_values = []
    output_scheduled_basal_rates = []
    output_delivered_units = []

    for i in range(0, len(dose_types)):
        (dose_type,
         start_date,
         end_date,
         value,
         scheduled_basal_rate,
         delivered_unit
         ) = annotate_individual_dose(
             dose_types[i], start_dates[i], end_dates[i], values[i], delivered_units[i],
             basal_start_times, basal_rates, basal_minutes,
             convert_to_units_hr
             )

        output_types.extend(dose_type)
        output_start_dates.extend(start_date)
        output_end_dates.extend(end_date)
        output_values.extend(value)
        output_scheduled_basal_rates.extend(scheduled_basal_rate)
        output_delivered_units.extend(delivered_unit)

    assert len(output_types) == len(output_start_dates) ==\
        len(output_end_dates) == len(output_values) ==\
        len(output_scheduled_basal_rates) == len(output_delivered_units), "expected output shapes to match"

    return (output_types,
            output_start_dates,
            output_end_dates,
            output_values,
            output_scheduled_basal_rates,
            output_delivered_units
            )


def annotate_individual_dose(dose_type, dose_start_date, dose_end_date, value, delivered_unit,
                             basal_start_times, basal_rates, basal_minutes,
                             convert_to_units_hr=True):
    """ Annotates a dose with the context of the scheduled basal rate
        If the dose crosses a schedule boundary, it will be split into
        multiple doses so each dose has a single scheduled basal rate.

    Arguments:
    dose_type -- type of dose (basal, bolus, etc)
    dose_start_date -- start date of the dose (datetime obj)
    dose_end_date -- end date of the dose (datetime obj)
    value -- actual basal rate of dose in U/hr (if a basal)
             or the value of the bolus in U
    basal_start_times -- list of times the basal rates start at
    basal_rates -- list of basal rates(U/hr)
    basal_minutes -- list of basal lengths (in mins)
    convert_to_units_hr -- set to True if you want to convert a dose to U/hr
        (ex: 0.05 U given from 1/1/01 1:00:00 to 1/1/01 1:05:00 -> 0.6 U/hr)

    Output:
    Tuple with properties of doses, annotated with the current basal rates
    """
    if dose_type not in [DoseType.basal, DoseType.tempbasal,
                                 DoseType.suspend]:
        return ([dose_type], [dose_start_date], [dose_end_date], [value],
                [0], [delivered_unit])

    output_types = []
    output_start_dates = []
    output_end_dates = []
    output_values = []
    output_scheduled_basal_rates = []
    output_delivered_units = []

    # these are the lists containing the scheduled basal value(s) within
    # the temp basal's duration
    (sched_basal_starts,
     sched_basal_ends,
     sched_basal_rates
     ) = between(
         basal_start_times,
         basal_rates,
         basal_minutes,
         dose_start_date,
         dose_end_date,
         )

    for i in range(0, len(sched_basal_starts)):
        if i == 0:
            start_date = dose_start_date
        else:
            start_date = sched_basal_starts[i]

        if i == len(sched_basal_starts) - 1:
            end_date = dose_end_date
        else:
            end_date = sched_basal_starts[i+1]

        output_types.append(dose_type)
        output_start_dates.append(start_date)
        output_end_dates.append(end_date)

        if convert_to_units_hr:
            output_values.append(
                0 if dose_type == DoseType.suspend else value
                / (time_interval_since(dose_end_date, dose_start_date)/60/60))
        else:
            output_values.append(value)

        output_scheduled_basal_rates.append(sched_basal_rates[i])
        output_delivered_units.append(delivered_unit)

    assert len(output_types) == len(output_start_dates) ==\
        len(output_end_dates) == len(output_values) ==\
        len(output_scheduled_basal_rates) == len(output_delivered_units), "expected output shapes to match"

    return (output_types,
            output_start_dates,
            output_end_dates,
            output_values,
            output_scheduled_basal_rates,
            output_delivered_units
            )


def between(
        basal_start_times, basal_rates, basal_minutes,
        start_date, end_date,
        repeat_interval=24
    ):
    """ Returns a slice of scheduled basal rates that occur between two dates

    Arguments:
    basal_start_times -- list of times the basal rates start at
    basal_rates -- list of basal rates(U/hr)
    basal_minutes -- list of basal lengths (in mins)
    start_date -- start date of the range (datetime obj)
    end_date -- end date of the range (datetime obj)
    repeat_interval -- the duration over which the rates repeat themselves
                       (24 hours by default)

    Output:
    Tuple in format (basal_start_times, basal_rates, basal_minutes) within
    the range of dose_start_date and dose_end_date
    """
    timezone_info = start_date.tzinfo
    if start_date > end_date:
        return ([], [], [])

    reference_time_interval = timedelta(
        hours=basal_start_times[0].hour,
        minutes=basal_start_times[0].minute,
        seconds=basal_start_times[0].second
    )
    max_time_interval = (
        reference_time_interval
        + timedelta(hours=repeat_interval)
    )

    start_offset = schedule_offset(start_date, basal_start_times[0])

    end_offset = (
        start_offset
        + timedelta(seconds=time_interval_since(end_date, start_date))
    )

    # if a dose is crosses days, split it into separate doses
    if end_offset > max_time_interval:
        boundary_date = start_date + (max_time_interval - start_offset)
        (start_times_1,
         end_times_1,
         basal_rates_1
         ) = between(
             basal_start_times,
             basal_rates,
             basal_minutes,
             start_date,
             boundary_date,
             repeat_interval=repeat_interval
             )
        (start_times_2,
         end_times_2,
         basal_rates_2
         ) = between(
             basal_start_times,
             basal_rates,
             basal_minutes,
             boundary_date,
             end_date,
             repeat_interval=repeat_interval
             )

        return (start_times_1 + start_times_2,
                end_times_1 + end_times_2,
                basal_rates_1 + basal_rates_2
                )

    start_index = 0
    end_index = len(basal_start_times)

    for (i, start_time) in enumerate(basal_start_times):
        start_time = timedelta(
            hours=start_time.hour,
            minutes=start_time.minute,
            seconds=start_time.second
        )
        if start_offset >= start_time:
            start_index = i
        if end_offset < start_time:
            end_index = i
            break

    reference_date = start_date - start_offset
    reference_date = datetime(
        year=reference_date.year,
        month=reference_date.month,
        day=reference_date.day,
        hour=reference_date.hour,
        minute=reference_date.minute,
        second=reference_date.second,
        tzinfo=timezone_info
        )

    if start_index > end_index:
        return ([], [], [])

    (output_start_times, output_end_times, output_basal_rates) = ([], [], [])

    for i in range(start_index, end_index):
        end_time = (timedelta(
            hours=basal_start_times[i+1].hour,
            minutes=basal_start_times[i+1].minute,
            seconds=basal_start_times[i+1].second) if i+1 <
                    len(basal_start_times) else max_time_interval)

        output_start_times.append(
            reference_date + timedelta(
                hours=basal_start_times[i].hour,
                minutes=basal_start_times[i].minute,
                seconds=basal_start_times[i].second
            )
        )
        output_end_times.append(reference_date + end_time)
        output_basal_rates.append(basal_rates[i])

    assert len(output_start_times) == len(output_end_times) ==\
        len(output_basal_rates), "expected output shape to match"

    return (output_start_times, output_end_times, output_basal_rates)


def schedule_offset(date_to_offset, reference_time,
                    repeat_interval=24):
    """ Returns the time interval for a given date normalized to the span of
        the schedule items

    Arguments:
    date_to_offset -- datetime object of the date to convert
    reference_time -- time object that's the first basal dose time in
                      a basal schedule (typically midnight)
    repeat_interval -- the interval with which the basal schedule repeats

    Output:
    datetime timedelta object representing offset
    """
    reference_time_seconds = (reference_time.hour * 3600
                              + reference_time.minute * 60
                              + reference_time.second)
    interval = time_interval_since_reference_date(date_to_offset)

    return timedelta(seconds=(interval - reference_time_seconds)
                     % (repeat_interval * 60 * 60)
                     + reference_time_seconds)


def insulin_on_board(
        dose_types, start_dates, end_dates, values, scheduled_basal_rates, delivered_units,
        model,
        start=None,
        end=None,
        delay=10,
        delta=5
    ):
    """ Calculates the timeline of insulin remaining for a collection of doses

        This model allows us to specify time of peak activity, as well as
        duration, and provides activity and IOB decay functions

    Arguments:
    dose_types -- list of types of doses (basal, bolus, etc)
    start_dates -- list of datetime objects representing the dates
                   the doses started at
    end_dates -- list of datetime objects representing the dates
                   the doses ended at
    values -- list of insulin values for doses
    scheduled_basal_rates -- basal rates scheduled during the times of doses
    delivered_units -- units actually delivered by dose
    model -- list of insulin model parameters in format [DIA, peak_time]
    start -- datetime object of time to start calculating the IOB timeline
    end -- datetime object of time to end the IOB timeline
    delay -- the time to delay the dose effect
    delta -- the differential between timeline entries

    Output:
    Tuple in format (times_iob_was_calculated_at, iob_values (U of insulin))
    """
    assert len(dose_types) == len(start_dates) == len(end_dates) ==\
        len(values) == len(scheduled_basal_rates) == len(delivered_units),\
        "expected input shapes to match"

    if not dose_types:
        return ([], [])

    try:
        if len(model) == 1:
            (start, end
             ) = simulation_date_range_for_samples(
                 start_times=start_dates,
                 end_times=end_dates,
                 duration=model[0]*60,
                 delay=delay,
                 delta=delta
                 )
        else:
            (start, end
             ) = simulation_date_range_for_samples(
                 start_times=start_dates,
                 end_times=end_dates,
                 duration=model[0],
                 delay=delay,
                 delta=delta
                 )
    except IndexError:
        return ([], [])

    date = start
    iob_dates = []
    iob_values = []

    def find_partial_iob(i):
        return insulin_on_board_calc(
            dose_types[i],
            start_dates[i],
            end_dates[i],
            values[i],
            scheduled_basal_rates[i],
            delivered_units[i],
            date,
            model,
            delay,
            delta
            )

    while date <= end:
        iob_sum = 0
        for i in range(0, len(start_dates)):
            iob_sum += find_partial_iob(i)

        iob_dates.append(date)
        iob_values.append(iob_sum)
        date += timedelta(minutes=delta)

    assert len(iob_dates) == len(iob_values), "expected output shape to match"

    return (iob_dates, iob_values)


def insulin_on_board_calc(
        type_, start_date, end_date, value, scheduled_basal_rate, delivered_units,
        date,
        model,
        delay,
        delta
    ):
    """ Calculates the insulin on board for a specific dose at a specific time

    Arguments:
    type_ -- String with type of dose (bolus, basal, etc)
    start_date -- the date the dose started at (datetime object)
    end_date -- the date the dose ended at (datetime object)
    value -- insulin value for dose
    scheduled_basal_rate -- basal rate scheduled during the times of dose
                            (0 for a bolus)
    delivered_units -- units actually delivered by pump
    date -- date the IOB is being calculated (datetime object)
    model -- list of insulin model parameters in format [DIA, peak_time]
    delay -- the time to delay the dose effect
    delta -- the differential between timeline entries

    Output:
    IOB at date
    """
    time = time_interval_since(date, start_date)

    if start_date > end_date or time < 0:
        return 0

    if len(model) == 1:  # walsh model
        if time_interval_since(end_date, start_date) <= 1.05 * delta * 60:
            return net_basal_units(
                type_,
                value,
                start_date,
                end_date,
                scheduled_basal_rate,
                delivered_units,
                ) * walsh_percent_effect_remaining(
                    (time / 60 - delay),
                    model[0]
                    )
        # This will normally be for basals
        return net_basal_units(
            type_,
            value,
            start_date,
            end_date,
            scheduled_basal_rate,
            delivered_units) * continuous_delivery_insulin_on_board(
                start_date,
                end_date,
                date,
                model,
                delay,
                delta
                )

    # Consider doses within the delta time window as momentary
    # This will normally be for boluses or short temp basals
    if time_interval_since(end_date, start_date) <= 1.05 * delta * 60:
        return net_basal_units(
            type_,
            value,
            start_date,
            end_date,
            scheduled_basal_rate,
            delivered_units
            ) * percent_effect_remaining(
                (time / 60 - delay),
                model[0],
                model[1]
                )
    # This will normally be for basals
    return net_basal_units(
        type_,
        value,
        start_date,
        end_date,
        scheduled_basal_rate,
        delivered_units) * continuous_delivery_insulin_on_board(
            start_date,
            end_date,
            date,
            model,
            delay,
            delta
            )


def continuous_delivery_insulin_on_board(
        start_date,
        end_date,
        at_date,
        model,
        delay,
        delta
    ):
    """ Calculates the percent of original insulin that is still on board
         at a specific time for a dose given over a period greater than
         1.05x the delta (this will almost always be a basal)

    Arguments:
    start_date -- the date the dose started at (datetime object)
    end_date -- the date the dose ended at (datetime object)
    at_date -- date the IOB is being calculated (datetime object)
    model -- list of insulin model parameters in format [DIA, peak_time]
    delay -- the time to delay the dose effect
    delta -- the differential between timeline entries

    Output:
    Percentage of insulin remaining at the at_date
    """
    dose_duration = time_interval_since(end_date, start_date)
    delay *= 60
    delta *= 60

    if dose_duration < 0:
        return 0

    time = time_interval_since(at_date, start_date)
    iob = 0
    dose_date = 0

    while (dose_date <= min(floor((time + delay) / delta) * delta,
                            dose_duration)):
        if dose_duration > 0:
            segment = (max(
                0,
                min(dose_date + delta,
                    dose_duration
                    )
                - dose_date) / dose_duration)
        else:
            segment = 1
        if len(model) == 1:  # if walsh model
            iob += segment * walsh_percent_effect_remaining(
                (time - delay - dose_date) / 60,
                model[0]
                )
        else:
            iob += segment * percent_effect_remaining(
                (time - delay - dose_date) / 60,
                model[0],
                model[1]
                )
        dose_date += delta

    return iob


def glucose_effects(
        dose_types,
        dose_start_dates,
        dose_end_dates,
        dose_values,
        scheduled_basal_rates,
        delivered_units,
        model,
        sensitivity_start_times,
        sensitivity_end_times,
        sensitivity_values,
        delay=10,
        delta=5,
        start=None,
        end=None
        ):
    """ Calculates the timeline of glucose effects for a collection of doses

    Arguments:
    dose_types -- list of types of doses (basal, bolus, etc)
    dose_start_dates -- list of datetime objects representing the dates
                       the doses started at
    dose_end_dates -- list of datetime objects representing the dates
                       the doses ended at
    dose_values -- list of insulin values for doses
    scheduled_basal_rates -- basal rates scheduled during the times of doses

    model -- list of insulin model parameters in format [DIA, peak_time] if
             exponential model, or [DIA] if Walsh model

    sensitivity_start_times -- list of time objects of start times of
                               given insulin sensitivity values
    sensitivity_end_times -- list of time objects of start times of
                             given insulin sensitivity values
    sensitivity_values -- list of sensitivities (mg/dL/U)

    delay -- the time to delay the dose effect
    delta -- the differential between timeline entries

    start -- datetime to start calculating the effects at
    end -- datetime to end calculation of effects

    Output:
    Tuple in format (times_glucose_effect_was_calculated_at,
                     glucose_effect_values (mg/dL))
    """
    assert len(dose_types) == len(dose_start_dates) == len(dose_end_dates)\
        == len(dose_values) == len(scheduled_basal_rates) == len(delivered_units),\
        "expected input shapes to match"

    if not dose_types and not (start is not None and end is not None):
        return ([], [])

    if len(model) == 1:  # if using a Walsh model
        start, end = simulation_date_range_for_samples(
            start_times=dose_start_dates,
            end_times=dose_end_dates,
            duration=model[0] * 60,
            delay=delay,
            delta=delta,
            start=start,
            end=end
        )
    else:
        start, end = simulation_date_range_for_samples(
            start_times=dose_start_dates,
            end_times=dose_end_dates,
            duration=model[0],
            delay=delay,
            delta=delta,
            start=start,
            end=end
        )

    date = start
    effect_dates = []
    effect_values = []

    def find_partial_effect(i):
        sensitivity = find_ratio_at_time(
            sensitivity_start_times,
            sensitivity_end_times,
            sensitivity_values,
            dose_start_dates[i]
        )
        return glucose_effect(
            dose_types[i],
            dose_start_dates[i],
            dose_end_dates[i],
            dose_values[i],
            scheduled_basal_rates[i],
            delivered_units[i],
            date,
            model,
            sensitivity,
            delay,
            delta
        )

    while date <= end:
        effect_sum = 0
        for i in range(0, len(dose_start_dates)):
            effect_sum += find_partial_effect(i)

        effect_dates.append(date)
        effect_values.append(effect_sum)
        date += timedelta(minutes=delta)

    assert len(effect_dates) == len(effect_values),\
        "expected output shapes to match"
    return (effect_dates, effect_values)


def find_ratio_at_time(ratio_start_times, ratio_end_times,
                       ratio_values, time_to_check
                       ):
    """ Finds ratio or correction range value at a given time

    Arguments:
    ratio_start_times -- list of time objects of start times of
                               given ratio/correction range values
    ratio_end_times -- list of time objects of start times of
                             given ratio/correction range values
    ratio_values -- list of ratio/correction range values
    time_to_check -- finding the value at this (date)time

    Output:
    Value at time_to_check
    """

    assert len(ratio_start_times) == len(ratio_values),\
        "expected input shapes to match"

    for i in range(0, len(ratio_start_times)):
        if ratio_end_times:
            if is_time_between(
                    ratio_start_times[i],
                    ratio_end_times[i],
                    time_to_check
                    ):  # pylint: disable=C0330
                return ratio_values[i]
        else:
            if is_time_between(
                    ratio_start_times[i],
                    (ratio_start_times[i+1]
                     if i+1 < len(ratio_start_times)
                     else ratio_start_times[0]
                    ),
                    time_to_check
                    ):  # pylint: disable=C0330
                return ratio_values[i]
    return 0


def is_time_between(start, end, time_to_check):
    """ Check if time is within an interval

    Arguments:
    start -- time (or datetime) of start of interval
    end -- time (or datetime) of end of interval
    time_to_check -- see if this time (or datetime) value is within the
                     interval

    Output:
    True if within interval, False if not
    """
    # convert from datetime to time if needed so we can compare
    if isinstance(start, datetime):
        start = start.time()
    if isinstance(end, datetime):
        end = end.time()
    if isinstance(time_to_check, datetime):
        time_to_check = time_to_check.time()

    if start < end:
        return start <= time_to_check <= end
    # if it crosses midnight
    return time_to_check >= start or time_to_check <= end


def glucose_effect(
        dose_type,
        dose_start_date,
        dose_end_date,
        dose_value,
        scheduled_basal_rate,
        delivered_units,
        date,
        model,
        insulin_sensitivity,
        delay,
        delta
    ):
    """ Calculates the timeline of glucose effects for a specific dose

    Arguments:
    dose_type -- types of dose (basal, bolus, etc)
    dose_start_date -- datetime object representing date doses start at
    dose_end_date -- datetime object representing date dose ended at
    dose_value -- insulin value for dose
    scheduled_basal_rate -- basal rate scheduled during the time of dose
    date -- datetime object of time to calculate the effect at
    insulin_sensitivity -- sensitivity (mg/dL/U)
    delay -- the time to delay the dose effect
    delta -- the differential between timeline entries

    Output:
    Glucose effect (mg/dL)
    """
    time = time_interval_since(date, dose_start_date)
    delay *= 60
    delta *= 60

    if time < 0:
        return 0

    # Consider doses within the delta time window as momentary
    # This will normally be for boluses
    if time_interval_since(
                dose_end_date,
                dose_start_date
            ) <= 1.05 * delta:  # pylint: disable=C0330

        if len(model) == 1:  # walsh model
            return net_basal_units(
                dose_type, dose_value,
                dose_start_date,
                dose_end_date,
                scheduled_basal_rate,
                delivered_units
                ) * -insulin_sensitivity * (1 - walsh_percent_effect_remaining(
                    (time - delay) / 60,
                    model[0]
                    ))

        return net_basal_units(
            dose_type,
            dose_value,
            dose_start_date,
            dose_end_date,
            scheduled_basal_rate,
            delivered_units
            ) * -insulin_sensitivity * (1 - percent_effect_remaining(
                (time - delay) / 60,
                model[0],
                model[1]
                ))
    # This will normally be for basals, and handles Walsh model automatically
    return net_basal_units(
        dose_type,
        dose_value,
        dose_start_date,
        dose_end_date,
        scheduled_basal_rate,
        delivered_units
        ) * -insulin_sensitivity * continuous_delivery_glucose_effect(
            dose_start_date,
            dose_end_date,
            date,
            model,
            delay / 60,
            delta / 60
            )


def continuous_delivery_glucose_effect(
        dose_start_date, dose_end_date,
        at_date,
        model,
        delay,
        delta
    ):
    """ Calculates the percent of glucose effect at a specific time for
        a dose given over a period greater than 1.05x the delta
        (this will almost always be a basal)

    Arguments:
    dose_start_date -- the date the dose started at (datetime object)
    dose_end_date -- the date the dose ended at (datetime object)
    at_date -- date the IOB is being calculated (datetime object)
    model -- list of insulin model parameters in format [DIA, peak_time]
    delay -- the time to delay the dose effect
    delta -- the differential between timeline entries

    Output:
    Percentage of insulin remaining at the at_date
    """
    dose_duration = time_interval_since(dose_end_date, dose_start_date)
    delay *= 60
    delta *= 60

    if dose_duration < 0:
        return 0

    time = time_interval_since(at_date, dose_start_date)
    activity = 0
    dose_date = 0

    while (dose_date <= min(
            floor((time + delay) / delta) * delta,
            dose_duration)):
        if dose_duration > 0:
            segment = (max(0,
                           min(dose_date + delta,
                               dose_duration
                               )
                           - dose_date
                           ) / dose_duration
                      )
        else:
            segment = 1

        if len(model) == 1:  # if walsh model
            activity += segment * (1 - walsh_percent_effect_remaining(
                (time - delay - dose_date) / 60,
                model[0])
                                  )

        else:
            activity += segment * (1 - percent_effect_remaining(
                (time - delay - dose_date) / 60,
                model[0],
                model[1]
                )
                                  )
        dose_date += delta

    return activity


def trim(
        dose_type, start, end, value, scheduled_basal_rate, delivered_units,
        start_interval=None,
        end_interval=None
    ):
    """ Trim doses to be within a particular interval

    Arguments:
    dose_type -- type of dose (basal, bolus, etc)
    start -- datetime object of time dose started at
    end -- datetime object of time dose ended at
    value -- amount, in U/hr (if a basal) or U (if bolus) of insulin in dose
    scheduled_basal_rate -- scheduled basal rate at the dose time
    start_interval -- start of interval to trim dose (datetime object)
    end_interval -- end of interval to trim dose (datetime object)

    Output:
    List with dose properties, trimmed to be in range (start_interval,
    end_interval). Format: [dose_type, dose_start_date, dose_end_date,
                            dose_value, dose_scheduled_rate]
    """
    if start.tzinfo:
        start_date = max(start_interval or DISTANT_PAST, start)
    else:
        start_date = max(start_interval or DISTANT_PAST, start)

    if end.tzinfo:
        return [dose_type,
                start_date,
                max(start_date,
                    min(end_interval or TIMEZONE_DISTANT_FUTURE, end)
                    ),
                value,
                scheduled_basal_rate,
                delivered_units
                ]

    return [dose_type,
            start_date,
            max(start_date, min(end_interval or DISTANT_FUTURE, end)),
            value,
            scheduled_basal_rate,
            delivered_units
            ]


def overlay_basal_schedule(
        dose_types, starts, ends, values,
        basal_start_times, basal_rates, basal_minutes,
        starting_at, ending_at,
        inserting_basal_entries
    ):
    """ Applies the current basal schedule to a collection of reconciled doses
        in chronological order

    Arguments:
    dose_types -- types of doses (basal, bolus, etc)
    starts -- datetime objects of times doses started at
    ends -- datetime objects of times doses ended at
    values -- amounts, in U/hr (if a basal) or U (if bolus) of insulin in doses
    basal_start_times -- list of times the basal rates start at
    basal_rates -- list of basal rates(U/hr)
    basal_minutes -- list of basal lengths (in mins)
    starting_at -- start of interval to overlay the basal schedule
                   (datetime object)
    ending_at -- end of interval to overlay the basal schedule
                 (datetime object)
    inserting_basal_entries -- whether basal doses should be created from the
                               schedule. Pass true only for pump models that do
                               not report their basal rates in event history.

    Output:
    Tuple with dose properties in range (start_interval,
    end_interval), overlayed with the basal schedule. It returns *four* dose
    properties, and does *not* return scheduled_basal_rates
    """
    assert len(dose_types) == len(starts) == len(ends) == len(values),\
        "expected input shapes to match"

    (out_dose_types, out_starts, out_ends, out_values) = ([], [], [], [])

    last_basal = []
    if inserting_basal_entries:
        last_basal = [DoseType.tempbasal,
                      starting_at,
                      starting_at,
                      0
                      ]

    for (i, type_) in enumerate(dose_types):
        if type_ in [DoseType.tempbasal, DoseType.basal, DoseType.suspend]:
            if ending_at and ends[i] > ending_at:
                continue

            if last_basal:
                if inserting_basal_entries:
                    (sched_basal_starts,
                     sched_basal_ends,
                     sched_basal_rates
                     ) = between(
                         basal_start_times,
                         basal_rates,
                         basal_minutes,
                         last_basal[2],
                         starts[i]
                         )
                    for j in range(0, len(sched_basal_starts)):
                        start = max(last_basal[2],
                                    sched_basal_starts[j]
                                    )
                        end = min(starts[i],
                                  sched_basal_ends[j]
                                  )

                        if time_interval_since(end, start)\
                                < sys.float_info.epsilon:
                            continue

                        out_dose_types.append(DoseType.basal)
                        out_starts.append(start)
                        out_ends.append(end)
                        out_values.append(sched_basal_rates[j])

            last_basal = [dose_types[i], starts[i], ends[i], values[i]]

            if last_basal:
                out_dose_types.append(last_basal[0])
                out_starts.append(last_basal[1])
                out_ends.append(last_basal[2])
                out_values.append(last_basal[3])

        elif type_ == DoseType.resume:
            assert "No resume events should be present in reconciled doses"

        elif type_ == DoseType.bolus:
            out_dose_types.append(dose_types[i])
            out_starts.append(starts[i])
            out_ends.append(ends[i])
            out_values.append(values[i])

    assert len(out_dose_types) == len(out_starts) == len(out_ends)\
        == len(out_values), "expected output shape to match"

    return (out_dose_types, out_starts, out_ends, out_values)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jul 12 13:35:43 2019

@author: annaquinlan
"""
# pylint: disable=C0200, C0103, R0912, R0913, R0914, R0915
import json
import os
import warnings

from datetime import datetime, time, timedelta
import numpy

from pyloopkit.dose import DoseType
from pyloopkit.loop_data_manager import update
from pyloopkit.loop_math import sort_dose_lists


# %% Functions to get various data from an issue report
def get_glucose_data(glucose_dict, offset=0):
    """ Load glucose values from an issue report cached_glucose dictionary

    Arguments:
    glucose_dict -- dictionary containing cached glucose sample information
    offset -- the offset from UTC in seconds

    Output:
    2 lists in (date, glucose_value) format
    """
    dates = [
        datetime.strptime(
            dict_.get("startDate"),
            "%Y-%m-%d %H:%M:%S %z"
        ) + timedelta(seconds=offset)
        for dict_ in glucose_dict
    ]

    glucose_values = [float(dict_.get("quantity")) for dict_ in glucose_dict]

    assert len(dates) == len(glucose_values),\
        "expected output shape to match"

    return (dates, glucose_values)


def convert_to_correct_units(type_, start, end, value):
    """ Take a dose and convert it into the appropriate unit
        (either U or U/hr)
    """
    if type_ == DoseType.bolus:
        return value

    return value / ((end - start).total_seconds()/60/60)


def get_insulin_data(
        data, offset=0, convert_to_units=False, entry_to_add=None,
        now_time=None):
    """ Load doses from an issue report cached_doses
        or normalized_insulin_doses dictionary

    Arguments:
    data -- dictionary containing cached dose information
    offset -- the offset from UTC in seconds
    convert_to_units -- convert from dose amounts to doses with the correct
                        units (so U for boluses and U/hr for basals)
    entry_to_add -- the last entry to add; this is normally used when getting
                    data from the "get_normalized_pump_event_dose" dictionary,
                    as it sometimes lacks the last insulin dose. This function
                    assumes that if it is a basal, it will need to be converted
                    to U/hr.
    now_time -- the time to run the loop at ("datetime.now")

    Output:
    4 lists in (dose_type (DoseType enum), start_dates, end_dates,
                values (in units/insulin)) format
    """
    dose_types = [
        DoseType.from_str(
            dict_.get("type")[17:]
        )
        if dict_.get("type").startswith("LoopKit.DoseType.")
        else DoseType.from_str(
            dict_.get("type")
        ) for dict_ in data
    ]
    start_dates = [
        datetime.strptime(
            dict_.get("startDate"),
            "%Y-%m-%d %H:%M:%S %z"
        ) + timedelta(seconds=offset)
        for dict_ in data
    ]
    end_dates = [
        datetime.strptime(
            dict_.get("endDate"),
            "%Y-%m-%d %H:%M:%S %z"
        ) + timedelta(seconds=offset)
        for dict_ in data
    ]
    values = []
    for i in range(0, len(data)):
        if 'deliveredUnits' in data[i].keys() and data[i].get("deliveredUnits") != 'nil':
            values.append(float(data[i].get("deliveredUnits")))
        else:
            values.append(
                convert_to_correct_units(
                    dose_types[i],
                    start_dates[i],
                    end_dates[i],
                    float(data[i].get("value")))
                if convert_to_units else float(data[i].get("value"))
            )

    if entry_to_add and now_time:
        start = datetime.strptime(
            entry_to_add.get("startDate"),
            "%Y-%m-%d %H:%M:%S %z"
        ) + timedelta(seconds=offset)
        dose_end = datetime.strptime(
            entry_to_add.get("endDate"),
            "%Y-%m-%d %H:%M:%S %z"
        ) + timedelta(seconds=offset)

        # if this entry is truly a new entry, convert to the appropriate units
        # and add to the output
        if not start == start_dates[-1] and not dose_end == end_dates[-1]:
            dose_types.append(
                DoseType.from_str(
                    entry_to_add.get("type")[17:]
                )
                if entry_to_add.get("type").startswith("LoopKit.DoseType.")
                else DoseType.from_str(
                    entry_to_add.get("type")
                )
            )
            start_dates.append(start)
            end_dates.append(
                max(
                    start,
                    min(now_time, dose_end)
                )
            )
            values.append(
                float(entry_to_add.get("value"))
            )

    assert len(dose_types) == len(start_dates) == len(end_dates) ==\
        len(values),\
        "expected output shape to match"

    return (dose_types, start_dates, end_dates, values)


def get_carb_data(data, offset=0):
    """ Load carb information from an issue report cached_carbs dictionary

    Arguments:
    data -- dictionary containing cached carb information
    offset -- the offset from UTC in seconds

    Output:
    3 lists in (carb_values, carb_start_dates, carb_absorption_times)
    format
    """
    carb_values = [float(dict_.get("quantity")) for dict_ in data]
    start_dates = [
        datetime.strptime(
            dict_.get("startDate"),
            " %Y-%m-%d %H:%M:%S %z"
        ) + timedelta(seconds=offset)
        for dict_ in data
    ]
    absorption_times = [
        float(dict_.get("absorptionTime")) / 60
        if dict_.get("absorptionTime") is not None
        else None for dict_ in data
    ]

    assert len(start_dates) == len(carb_values) == len(absorption_times),\
        "expected input shapes to match"
    return (start_dates, carb_values, absorption_times)


def seconds_to_time(seconds):
    """ Convert seconds since midnight into a datetime.time object """
    hours, remainder = divmod(seconds, 3600)
    minutes, seconds = divmod(remainder, 60)

    return time(int(hours), int(minutes), int(seconds))


def get_starts_and_ends_from_seconds(seconds_list):
    """ Given a list of seconds since midnight,
        convert into start and end times
    """
    starts = [seconds_to_time(seconds) for seconds in seconds_list]
    ends = [value for value in starts]
    ends.append(ends.pop(0))

    assert len(starts) == len(ends), "expected output shapes to match"

    return (starts, ends)


def get_sensitivities(data):
    """ Load insulin sensitivity schedule
        from an issue report isf_schedule dictionary

    Arguments:
    data -- dictionary containing ISF information

    Output:
    3 lists in (sensitivity_start_time, sensitivity_end_time,
                sensitivity_value (mg/dL/U)) format
    """
    seconds = [float(dict_.get("startTime")) for dict_ in data]

    (start_times, end_times) = get_starts_and_ends_from_seconds(seconds)

    values = [dict_.get("value") for dict_ in data]

    assert len(start_times) == len(end_times) == len(values),\
        "expected output shape to match"

    return (start_times, end_times, values)


def get_carb_ratios(data):
    """ Load carb ratio schedule
        from an issue report carb_ratio_schedule dictionary

    Arguments:
    data -- dictionary containing CR information

    Output:
    2 lists in (ratio_start_time, ratio_value (g/U)) format
    """
    seconds = [float(dict_.get("startTime")) for dict_ in data]

    start_times = get_starts_and_ends_from_seconds(seconds)[0]

    values = [dict_.get("value") for dict_ in data]

    assert len(start_times) == len(values),\
        "expected output shape to match"

    return (start_times, values)


def get_basal_schedule(data):
    """ Load basal rate schedule
        from an issue report basal_rate_schedule dictionary

    Arguments:
    data -- dictionary containing basal schedule information

    Output:
    3 lists in (rate_start_time, rate_length (minutes), rate (U/hr)) format
    """
    seconds = [float(dict_.get("startTime")) for dict_ in data]
    rate_minutes = []
    for i in range(0, len(seconds)):
        if i == len(seconds) - 1:
            rate_minutes.append(
                (seconds[i] - seconds[0]) / 60
            )
        else:
            rate_minutes.append(
                (seconds[i+1] - seconds[i]) / 60
            )

    start_times = get_starts_and_ends_from_seconds(seconds)[0]

    values = [dict_.get("value") for dict_ in data]

    assert len(start_times) == len(rate_minutes) == len(values),\
        "expected output shapes to match"

    return (start_times, values, rate_minutes)


def get_target_range_schedule(data):
    """ Load target range schedule
        from an issue report "correction_range_schedule" dictionary
    """
    seconds = [float(dict_.get("startTime")) for dict_ in data]
    (start_times, end_times) = get_starts_and_ends_from_seconds(seconds)
    min_values = [float(dict_.get("value")[0]) for dict_ in data]
    max_values = [float(dict_.get("value")[1]) for dict_ in data]

    return (start_times, end_times, min_values, max_values)


def load_momentum_effects(data, offset=0):
    """ Load glucose momentum effects from a list """
    start_times = [
        datetime.strptime(
            dict_.get("startDate"),
            "%Y-%m-%d %H:%M:%S %z"
        ) + timedelta(seconds=offset)
        for dict_ in data
    ]
    values = [
        float(dict_.get("quantity")) for dict_ in data
    ]
    return (start_times, values)


def get_counteractions(data, offset=0):
    """ Load counteraction effect data from a list """
    start_times = [
        datetime.strptime(
            dict_.get("start_time"),
            "%Y-%m-%d %H:%M:%S %z"
        ) + timedelta(seconds=offset)
        for dict_ in data
    ]
    end_times = [
        datetime.strptime(
            dict_.get("end_time"),
            " %Y-%m-%d %H:%M:%S %z"
        ) + timedelta(seconds=offset)
        for dict_ in data
    ]
    values = [
        float(dict_.get("value")) for dict_ in data
    ]
    return (start_times, end_times, values)


def load_insulin_effects(data, offset=0):
    """ Load insulin effect data from a list """
    start_times = [
        datetime.strptime(
            dict_.get("start_time"),
            "%Y-%m-%d %H:%M:%S %z"
        ) + timedelta(seconds=offset)
        for dict_ in data
    ]
    values = [
        float(dict_.get("value")) for dict_ in data
    ]
    return (start_times, values)


def get_retrospective_effects(data, offset=0):
    """ Load retrospective effect data from a list """
    start_times = [
        datetime.strptime(
            dict_.get("startDate"),
            "%Y-%m-%d %H:%M:%S %z"
        ) + timedelta(seconds=offset)
        for dict_ in data
    ]
    values = [
        float(dict_.get("quantity")) for dict_ in data
    ]
    return (start_times, values)


def get_settings(data):
    """ Load needed settings from an issue report

    Arguments:
    data -- the parsed issue report dictionary

    Output:
    Dictionary of settings
    """
    settings = {}

    model = data.get("insulin_model")
    if not model:
        raise RuntimeError("No insulin model information found")

    if model.lower() == "humalognovologchild":
        settings["model"] = [
            data.get("insulin_action_duration") / 60,
            65
        ]
    elif model.lower() == "humalognovologadult":
        settings["model"] = [
            data.get("insulin_action_duration") / 60,
            75
        ]
    elif model.lower() == "fiasp":
        settings["model"] = [
            data.get("insulin_action_duration") / 60,
            55
        ]
    else:  # Walsh model
        settings["model"] = [
            data.get("insulin_action_duration") / 60 / 60
        ]

    momentum_interval = data.get("glucose_store").get("momentumDataInterval")
    if momentum_interval is not None:
        settings["momentum_data_interval"] = float(momentum_interval) / 60
    else:
        settings["momentum_data_interval"] = 15

    suspend_threshold = data.get("suspend_threshold")
    if suspend_threshold is not None:
        settings["suspend_threshold"] = float(suspend_threshold)
    else:
        settings["suspend_threshold"] = None

    settings["dynamic_carb_absorption_enabled"] = True
    settings["retrospective_correction_integration_interval"] = 30
    settings["recency_interval"] = 15
    settings["retrospective_correction_grouping_interval"] = 30
    settings["rate_rounder"] = 0.05
    settings["insulin_delay"] = 10
    settings["carb_delay"] = 10

    settings["default_absorption_times"] = [
        float(data.get("carb_default_absorption_times_fast")) / 60,
        float(data.get("carb_default_absorption_times_medium")) / 60,
        float(data.get("carb_default_absorption_times_slow")) / 60
        ]

    settings["max_basal_rate"] = data.get("maximum_basal_rate")
    settings["max_bolus"] = data.get("maximum_bolus")
    settings["retrospective_correction_enabled"] = data.get(
        "retrospective_correction_enabled"
    ) and data.get(
        "retrospective_correction_enabled"
    ).lower() == "true"

    return settings


def get_last_temp_basal(data, offset=0):
    """ Load the last temporary basal from an issue report
        "last_temp_basal" dictionary
    """
    if (data.get(" type") == "LoopKit.DoseType.tempBasal"
            or data.get("type") == "LoopKit.DoseType.tempBasal"):
        type_ = DoseType.tempbasal
    elif (data.get(" type") == "LoopKit.DoseType.basal"
          or data.get("type") == "LoopKit.DoseType.basal"):
        type_ = DoseType.basal
    else:
        raise RuntimeError("The last temporary basal is not a basal")

    return [
        type_,
        datetime.strptime(
            data.get(" startDate") if data.get(" startDate") is not None
            else data.get("startDate"),
            "%Y-%m-%d %H:%M:%S %z"
        ) + timedelta(seconds=offset),
        datetime.strptime(
            data.get(" endDate") if data.get(" endDate") is not None
            else data.get("endDate"),
            "%Y-%m-%d %H:%M:%S %z"
        ) + timedelta(seconds=offset),
        float(data.get(" value")) if data.get(" value") is not None
        else float(data.get("value"))
    ]


# %% List management tools
def sort_by_first_list(list_1, list_2, list_3=None, list_4=None, list_5=None):
    """ Sort lists that are matched index-wise, using the first list as the
        property to sort by

    Example:
        l1: [50, 2, 3]               ->     [2, 3, 50]
        l2: [dog, cat, parrot]       ->     [cat, parrot, dog]
    """
    unsort_1 = numpy.array(list_1)
    unsort_2 = numpy.array(list_2)
    unsort_3 = numpy.array(list_3)
    unsort_4 = numpy.array(list_4)
    unsort_5 = numpy.array(list_5)

    sort_indexes = unsort_1.argsort()

    unsort_1.sort()
    list_1 = list(unsort_1)
    l2 = list(unsort_2[sort_indexes])
    if list_3:
        l3 = list(unsort_3[sort_indexes])
    else:
        l3 = []
    if list_4:
        l4 = list(unsort_4[sort_indexes])
    else:
        l4 = []
    if list_5:
        l5 = list(unsort_5[sort_indexes])
    else:
        l5 = []

    return (list_1, l2, l3, l4, l5)


def remove_too_new_values(
        sort_time,
        list_1, list_2, list_3=None, list_4=None, list_5=None,
        is_dose_data=False
        ):
    """ Remove values that occur after a certain date. This function makes the
        assumption that the date list is sorted in ascending order, and
        that all lists (if they are not None) are the same length. The first
        list must be the list with the times, unless is_dose_data is True,
        in which case the second list must contain the times.

    Arguments:
    sort_time -- the datetime after which to remove values
    """
    l1 = []
    l2 = []
    l3 = []
    l4 = []
    l5 = []

    for i in range(0, len(list_1)):
        # if this isn't dose data, use the first list to sort
        if not is_dose_data and list_1[i] <= sort_time:
            l1.append(list_1[i])
            l2.append(list_2[i])
            if list_3:
                l3.append(list_3[i])
            if list_4:
                l4.append(list_4[i])
            if list_5:
                l5.append(list_5[i])
        # otherwise, use the second list to sort
        elif is_dose_data and list_2[i] <= sort_time:
            l1.append(list_1[i])
            l2.append(list_2[i])
            if list_3:
                l3.append(list_3[i])
            if list_4:
                l4.append(list_4[i])
            if list_5:
                l5.append(list_5[i])

    return (l1, l2, l3, l4, l5)


# %% Take an issue report and run it through the Loop algorithm
def parse_report_and_run(path, name):
    return parse_report_and_run_with_name(os.path.join(path, name))

 # %% Take an issue report and run it through the Loop algorithm
def parse_report_and_run_with_name(data_path_and_name):
    """ Get relevent information from a Loop issue report and use it to
        run PyLoopKit

    Arguments:
    path -- the path to the issue report
    name -- the name of the file, with the .json extension

    Output:
    A dictionary of all 4 effects, the predicted glucose values, and the
    recommended basal and bolus
    """

    with open(data_path_and_name, "r") as file:
        issue_dict = json.load(file)
    input_dict = {}

    if issue_dict.get("basal_rate_timeZone") is not None:
        offset = issue_dict.get("basal_rate_timeZone")
    elif issue_dict.get("carb_ratio_timeZone") is not None:
        offset = issue_dict.get("carb_ratio_timeZone")
    elif issue_dict.get("insulin_sensitivity_factor_timeZone") is not None:
        offset = issue_dict.get("insulin_sensitivity_factor_timeZone")
    else:
        offset = 0
    input_dict["offset_applied_to_dates"] = offset

    if issue_dict.get("recommended_temp_basal"):
        time_to_run = datetime.strptime(
            issue_dict.get("recommended_temp_basal").get("date")
            or issue_dict.get("recommended_temp_basal").get(" date"),
            "%Y-%m-%d %H:%M:%S %z"
        ) + timedelta(seconds=offset)
    elif issue_dict.get("recommended_bolus"):
        time_to_run = datetime.strptime(
            issue_dict.get("recommended_bolus").get("date") or
            issue_dict.get("recommended_bolus").get(" date"),
            "%Y-%m-%d %H:%M:%S %z"
        ) + timedelta(seconds=offset)
    else:
        raise RuntimeError("No information found about report time")

    input_dict["time_to_calculate_at"] = time_to_run

    if issue_dict.get("cached_glucose_samples"):
        (glucose_dates, glucose_values) = get_glucose_data(
            issue_dict.get("cached_glucose_samples"),
            offset
        )
        (glucose_dates, glucose_values) = remove_too_new_values(
            time_to_run,
            *sort_by_first_list(
                glucose_dates, glucose_values
            )[0:2]
        )[0:2]
        input_dict["glucose_dates"] = glucose_dates
        input_dict["glucose_values"] = glucose_values
        input_dict["glucose_units"] = "mg/dL"

    else:
        raise RuntimeError("No glucose information found")

    if (issue_dict.get("get_normalized_pump_event_dose")
            and issue_dict.get("get_normalized_dose_entries")):
        (dose_types,
         dose_starts,
         dose_ends,
         dose_values
         ) = get_insulin_data(
            issue_dict.get("get_normalized_pump_event_dose"),
            offset,
            entry_to_add=issue_dict.get("get_normalized_dose_entries")[-1],
            now_time=time_to_run
        )
    elif issue_dict.get("get_normalized_pump_event_dose"):
        (dose_types,
         dose_starts,
         dose_ends,
         dose_values
         ) = get_insulin_data(
            issue_dict.get("get_normalized_pump_event_dose"),
            offset
        )
    elif issue_dict.get("get_normalized_dose_entries"):
        (dose_types,
         dose_starts,
         dose_ends,
         dose_values
         ) = get_insulin_data(
            issue_dict.get("get_normalized_dose_entries"),
            offset,
            convert_to_units=True
        )
    elif issue_dict.get("cached_dose_entries"):
        (dose_types,
         dose_starts,
         dose_ends,
         dose_values
         ) = get_insulin_data(
            issue_dict.get("cached_dose_entries"),
            offset,
            convert_to_units=True
        )
    else:
        warnings.warn("Warning: no insulin dose information found")
        (dose_types,
         dose_starts,
         dose_ends,
         dose_values
         ) = ([], [], [], [])

    (dose_types,
     dose_starts,
     dose_ends,
     dose_values
     ) = remove_too_new_values(
         time_to_run,
         *sort_dose_lists(
             dose_types,
             dose_starts,
             dose_ends,
             dose_values
         )[0:4],
         is_dose_data=True
    )[0:4]
    input_dict["dose_types"] = dose_types
    input_dict["dose_start_times"] = dose_starts
    input_dict["dose_end_times"] = dose_ends
    input_dict["dose_values"] = dose_values
    input_dict["dose_value_units"] = "U or U/hr"
    input_dict["dose_delivered_units"] = [None for i in range(len(dose_types))]

    if issue_dict.get("cached_carb_entries"):
        (carb_dates,
         carb_values,
         carb_absorptions
         ) = sort_by_first_list(
                 *get_carb_data(
                     issue_dict.get("cached_carb_entries"),
                     offset,
                 )
        )[0:3]
    else:
        (carb_dates,
         carb_values,
         carb_absorptions
         ) = ([], [], [])

    input_dict["carb_dates"] = carb_dates
    input_dict["carb_values"] = carb_values
    input_dict["carb_absorption_times"] = carb_absorptions
    input_dict["carb_value_units"] = "g"

    settings = get_settings(issue_dict)
    input_dict["settings_dictionary"] = settings

    if issue_dict.get(
            "insulinSensitivityScheduleApplyingOverrideHistory_items"):
        (sensitivity_start_times,
         sensitivity_end_times,
         sensitivity_values
         ) = get_sensitivities(
            issue_dict.get(
                "insulinSensitivityScheduleApplyingOverrideHistory_items"
            )
        )
    elif issue_dict.get("insulin_sensitivity_factor_schedule"):
        (sensitivity_start_times,
         sensitivity_end_times,
         sensitivity_values
         ) = get_sensitivities(
            issue_dict.get("insulin_sensitivity_factor_schedule")
        )
    else:
        raise RuntimeError("No insulin sensitivity information found")
    (sensitivity_start_times,
     sensitivity_end_times,
     sensitivity_values
     ) = sort_by_first_list(
         sensitivity_start_times,
         sensitivity_end_times,
         sensitivity_values
    )[0:3]

    input_dict["sensitivity_ratio_start_times"] = sensitivity_start_times
    input_dict["sensitivity_ratio_end_times"] = sensitivity_end_times
    input_dict["sensitivity_ratio_values"] = sensitivity_values
    input_dict["sensitivity_ratio_value_units"] = "mg/dL/U"

    if issue_dict.get("carbRatioScheduleApplyingOverrideHistory_items"):
        (carb_ratio_starts,
         carb_ratio_values
         ) = get_carb_ratios(
            issue_dict.get("carbRatioScheduleApplyingOverrideHistory_items")
        )
    elif issue_dict.get("carb_ratio_schedule"):
        (carb_ratio_starts,
         carb_ratio_values
         ) = get_carb_ratios(
            issue_dict.get("carb_ratio_schedule")
        )
    else:
        raise RuntimeError("No carb ratio information found")
    (carb_ratio_starts,
     carb_ratio_values
     ) = sort_by_first_list(
         carb_ratio_starts,
         carb_ratio_values
     )[0:2]

    input_dict["carb_ratio_start_times"] = carb_ratio_starts
    input_dict["carb_ratio_values"] = carb_ratio_values
    input_dict["carb_ratio_value_units"] = "g/U"

    if issue_dict.get("basalProfileApplyingOverrideHistory_items"):
        (basal_rate_starts,
         basal_rate_values,
         basal_rate_minutes
         ) = get_basal_schedule(
            issue_dict.get("basalProfileApplyingOverrideHistory_items")
        )
    elif issue_dict.get("basal_rate_schedule"):
        (basal_rate_starts,
         basal_rate_values,
         basal_rate_minutes
         ) = get_basal_schedule(
            issue_dict.get("basal_rate_schedule")
        )
    else:
        raise RuntimeError("No basal rate information found")
    (basal_rate_starts,
     basal_rate_minutes,
     basal_rate_values
     ) = sort_by_first_list(
         basal_rate_starts,
         basal_rate_minutes,
         basal_rate_values
     )[0:3]

    input_dict["basal_rate_start_times"] = basal_rate_starts
    input_dict["basal_rate_minutes"] = basal_rate_minutes
    input_dict["basal_rate_values"] = basal_rate_values
    input_dict["basal_rate_units"] = "U/hr"

    if issue_dict.get("correction_range_schedule"):
        (target_range_starts,
         target_range_ends,
         target_range_minimum_values,
         target_range_maximum_values
         ) = get_target_range_schedule(
            issue_dict.get("correction_range_schedule")
        )
        (target_range_starts,
         target_range_ends,
         target_range_minimum_values,
         target_range_maximum_values
         ) = sort_by_first_list(
             target_range_starts,
             target_range_ends,
             target_range_minimum_values,
             target_range_maximum_values
         )[0:4]
    else:
        raise RuntimeError("No target range rate information found")

    input_dict["target_range_start_times"] = target_range_starts
    input_dict["target_range_end_times"] = target_range_ends
    input_dict["target_range_minimum_values"] = target_range_minimum_values
    input_dict["target_range_maximum_values"] = target_range_maximum_values
    input_dict["target_range_value_units"] = "mg/dL"

    if issue_dict.get("last_temp_basal"):
        last_temp_basal = get_last_temp_basal(
            issue_dict.get("last_temp_basal"), offset
        )
    else:
        last_temp_basal = []
        warnings.warn(
            "No information found about the last temporary basal rate"
        )

    input_dict["last_temporary_basal"] = last_temp_basal

    recommendations = update(
        input_dict
        )

    return recommendations


def parse_dictionary_from_previous_run(path, name):
    """ Get a dictionary output from a previous run of PyLoopKit
        and convert the ISO strings to datetime or time objects, and
        dose types to enums
    """
    data_path_and_name = os.path.join(path, name)

    with open(data_path_and_name, "r") as file:
        dictionary = json.load(file)

    keys_with_times = [
        "basal_rate_start_times",
        "carb_ratio_start_times",
        "sensitivity_ratio_start_times",
        "sensitivity_ratio_end_times",
        "target_range_start_times",
        "target_range_end_times"
        ]

    for key in keys_with_times:
        new_list = []
        for string in dictionary.get(key):
            new_list.append(time.fromisoformat(string))
        dictionary[key] = new_list

    keys_with_datetimes = [
        "dose_start_times",
        "dose_end_times",
        "glucose_dates",
        "carb_dates"
        ]

    for key in keys_with_datetimes:
        new_list = []
        for string in dictionary.get(key):
            new_list.append(datetime.fromisoformat(string))
        dictionary[key] = new_list

    dictionary["time_to_calculate_at"] = datetime.fromisoformat(
        dictionary["time_to_calculate_at"]
    )

    last_temp = dictionary.get("last_temporary_basal")
    dictionary["last_temporary_basal"] = [
        DoseType.from_str(last_temp[0]),
        datetime.fromisoformat(last_temp[1]),
        datetime.fromisoformat(last_temp[2]),
        last_temp[3]
    ]

    dictionary["dose_types"] = [
        DoseType.from_str(value) for value in dictionary.get("dose_types")
    ]

    output = update(dictionary)

    return output



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jun 21 15:08:22 2019

@author: annaquinlan

Github URL: https://github.com/tidepool-org/LoopKit/blob/
57a9f2ba65ae3765ef7baafe66b883e654e08391/LoopKit/InsulinKit/
WalshInsulinModel.swift
"""


def walsh_percent_effect_remaining(minutes, action_duration):
    """ Give percent of insulin remaining for IOB calculations.
        This curve is only included for the purposes of running glucose
        activity tests

        Arguments:
        minutes -- minutes after insulin delivery
        dia -- duration of insulin action, in hours
    """
    if minutes <= 0:
        return 1
    if minutes >= action_duration * 60:
        return 0

    dia = round(action_duration)
    if dia < 3:
        dia = 3
    elif dia > 6:
        dia = 6

    minutes = minutes * dia / action_duration

    if dia == 3:
        return -3.2030e-9 * pow(minutes, 4) + 1.354e-6 * pow(minutes, 3)\
            - 1.759e-4 * pow(minutes, 2) + 9.255e-4 * minutes + 0.99951
    if dia == 4:
        return -3.310e-10 * pow(minutes, 4) + 2.530e-7 * pow(minutes, 3)\
            - 5.510e-5 * pow(minutes, 2) - 9.086e-4 * minutes + 0.99950
    if dia == 5:
        return -2.950e-10 * pow(minutes, 4) + 2.320e-7 * pow(minutes, 3)\
            - 5.550e-5 * pow(minutes, 2) + 4.490e-4 * minutes + 0.99300
    if dia == 6:
        return -1.493e-10 * pow(minutes, 4) + 1.413e-7 * pow(minutes, 3)\
            - 4.095e-5 * pow(minutes, 2) + 6.365e-4 * minutes + 0.99700

    raise RuntimeError



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jul 15 18:54:26 2019

@author: annaquinlan

Github URL: https://github.com/tidepool-org/LoopKit/blob/
57a9f2ba65ae3765ef7baafe66b883e654e08391/LoopKit/GlucoseKit/GlucoseStore.swift
"""
# pylint: disable=R0913, W0612
from datetime import timedelta

from pyloopkit.loop_math import filter_date_range
from pyloopkit.glucose_math import linear_momentum_effect, counteraction_effects


def get_recent_momentum_effects(
        glucose_starts, glucose_values,
        start_date,
        now_date,
        momentum_data_interval=15,
        delta=5,
        display_list=None,
        provenances=None
        ):
    """ Get glucose momentum effects

    Arguments:
    glucose_starts -- list of datetime objects of times of glucose values
    glucose_values -- list of glucose values (unit: mg/dL)

    start_date -- date to start calculating momentum effects
    now_date -- the date to assume as the "now" time (aka datetime.now())

    momentum_data_interval -- time to generate momentum effects out to (mins)
    delta -- time between blood glucose measurements (mins)

    display_list -- list of display_only booleans
    provenances -- list of provenances (Strings)

    Output:
    Momentum effects in format (date_of_effect, value_of_effect)
    """
    assert len(glucose_starts) == len(glucose_values),\
        "expected input shapes to match"

    if not glucose_starts or not start_date:
        return ([], [])

    (filtered_dates,
     ends,
     filtered_values) = filter_date_range(
         glucose_starts,
         [],
         glucose_values,
         now_date - timedelta(minutes=momentum_data_interval),
         None
         )

    if not display_list:
        display_list = [False for i in filtered_dates]
    if not provenances:
        provenances = ["PyLoop" for i in filtered_dates]

    effects = linear_momentum_effect(
        filtered_dates, filtered_values, display_list, provenances,
        momentum_data_interval,
        delta
        )

    return effects


def get_counteraction_effects(
        glucose_starts, glucose_values,
        start_date,
        effect_starts, effect_values,
        display_list=None,
        provenances=None
        ):
    """ Get counteraction effects

    Arguments:
    glucose_starts -- list of datetime objects of times of glucose values
    glucose_values -- list of glucose values (unit: mg/dL)

    start_date -- date to begin using glucose data (datetime)

    effect_dates -- list of datetime objects associated with a glucose effect
    effect_values -- list of values associated with a glucose effect

    display_list -- list of display_only booleans
    provenances -- list of provenances (Strings)

    Output:
    Counteraction effects in form (effect start, effect end, effect value)
    """
    assert len(glucose_starts) == len(glucose_values),\
        "expected input shapes to match"

    if not glucose_starts or not start_date:
        return ([], [])

    (filtered_starts,
     ends,
     filtered_values) = filter_date_range(
         glucose_starts,
         [],
         glucose_values,
         start_date,
         None
         )

    if not display_list:
        display_list = [False for i in filtered_starts]
    if not provenances:
        provenances = ["PyLoop" for i in filtered_starts]

    counteractions = counteraction_effects(
        filtered_starts, filtered_values, display_list, provenances,
        effect_starts, effect_values
        )

    return counteractions



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jul 29 16:39:23 2019

@author: annaquinlan
"""
# pylint: disable=R0911, W0613
import warnings
from pyloopkit.dose import DoseType


def are_settings_valid(settings):
    """ Checks that a settings dictionary has needed properties, and
        that the values of those properties are reasonable
        (no negative insulin peaks, etc)
    """
    model = settings.get("model")
    if (any(value <= 0 or value >= 1440 for value in model)
            or model[0] > 24 if len(model) == 1 else model[1] > 120
            or model[1] > model[0] if len(model) == 2 else False
       ):
        warnings.warn(
            "Error: expected insulin model with DIA between 0"
            + "and 24 hours, peak <= 120 mins, and peak < DIA; stopping run."
        )
        return False

    if (settings.get("momentum_data_interval")
            and settings.get("momentum_data_interval") < 5):
        warnings.warn(
            "Warning: momentum interval is less than 5"
            + " minutes; continuing anyway")

    if (settings.get("suspend_threshold")
            and settings.get("suspend_threshold") < 54
       ):
        warnings.warn(
            "Warning: suspend threshold < 54 mg/dL; continuing anyway"
        )
    elif (settings.get("suspend_threshold")
            and settings.get("suspend_threshold") > 180
       ):
        warnings.warn(
            "Warning: suspend threshold > 180 mg/dL; continuing anyway"
        )

    if any(absorption <= 0 for absorption in settings.get(
                "default_absorption_times")):
        warnings.warn(
            "Error: default absorption times must be positive; stopping run"
        )
        return False

    if (settings.get("max_basal_rate") < 0
            or settings.get("max_basal_rate") > 35
       ):
        warnings.warn(
            "Warning: maximum basal rate is typically at least 0 and less"
            + "than 35 U/hr; continuing anyway"
        )

    if (settings.get("max_bolus") < 0
            or settings.get("max_bolus") > 30
       ):
        warnings.warn(
            "Warning: maximum bolus is typically at least 0 and less than"
            + " 30 U; continuing anyway"
        )

    return True


def are_glucose_readings_valid(dates, glucose_values):
    """ Checks that glucose readings are reasonable """
    if any(value < 0 for value in glucose_values):
        warnings.warn(
            "Error: glucose measurements cannot be negative; stopping run"
        )
        return False

    if any(value < 39 or value > 400 for value in glucose_values):
        warnings.warn(
            "Warning: glucose measurements are typically between 39 and"
            + " 400 mg/dL; continuing anyway"
        )

    return True


def are_carb_readings_valid(dates, carb_values, absorption_times):
    """ Checks that carbohydrate inputs are reasonable """
    if any(value < 0 for value in carb_values):
        warnings.warn(
            "Error: carbohydrate value cannot be negative; stopping run."
        )
        return False

    if any(value > 250 for value in carb_values):
        warnings.warn(
            "Warning: data contains carbohydrate values > 250 g; continuing"
            + " anyway"
        )

    if any(
            absorption < 0 or absorption > 1440
            for absorption in absorption_times):
        warnings.warn(
            "Error: expected carbohydrate absorption times to be between"
            + " 0 & 1440 minutes (0 & 24 hours); stopping run"
        )
        return False

    return True


def are_insulin_doses_valid(types, start_times, end_times, values):
    """ Checks that dose inputs are reasonable """
    if any(type_ not in
           [DoseType.tempbasal, DoseType.basal, DoseType.bolus,
            DoseType.suspend, DoseType.resume, DoseType.meal
            ] for type_ in types):
        warnings.warn(
            "Warning: there are types in the insulin doses that PyLoopKit" +
            " doesn't recognize; continuing anyway"
        )

    if any(value < 0 or value > 35 for value in values):
        warnings.warn(
            "Warning: expected dose values to be between 0 and 35 U or U/hr;"
            + " continuing anyway"
        )

    if any(
            start > end for (start, end) in
            list(
                zip(
                    start_times, end_times
                    )
                )
            ):
        warnings.warn(
            "Error: dose start times cannot be greater than dose end times;"
            + " stopping run")
        return False

    return True


def is_insulin_sensitivity_schedule_valid(start_times, end_times, ratios):
    """ Checks that an insulin sensitivity schedule is reasonable """
    if any(value < 10 or value > 500 for value in ratios):
        warnings.warn(
            "Warning: data contains sensitivity values < 10 or > 500"
            + " mg/dL per Unit; continuing anyway"
            )

    if any(
            start > end for (start, end) in
            list(
                zip(
                    start_times, end_times
                    )
                )[:-1]):  # don't include the last entry because start > end
        warnings.warn(
            "Error: sensitivity ratio start times cannot be greater than ratio"
            + " end times; stopping run."
        )
        return False

    return True


def are_carb_ratios_valid(dates, ratios):
    """ Checks that carbohydrate ratios are reasonable """
    if any(value < 1 or value > 150 for value in ratios):
        warnings.warn(
            "Warning: data contains carb ratios < 1 or > 150 grams of carbs"
            + " per Unit; continuing anyway")

    return True


def are_basal_rates_valid(start_times, rates, minutes_active):
    """ Checks that scheduled basal rates are reasonable """
    if any(value < 0 for value in rates):
        warnings.warn(
            "Error: data contains negative scheduled basal rates; stopping run"
        )
        return False
    elif any(value > 35 for value in rates):
        warnings.warn(
            "Warning: data contains scheduled basal rates > 35 U/hr;"
            + " continuing anyway"
        )

    if any(duration > 1440 for duration in minutes_active):
        warnings.warn(
            "Error: data contains basal rates with scheduled duration greater"
            + " than a day (1440 mins); stopping run"
        )
        return False

    return True


def are_correction_ranges_valid(
        start_times, end_times, minimum_values, maximum_values):
    """ Checks that correction ranges are reasonable """
    if (any(value < 60 or value > 180 for value in minimum_values)
            or any(value < 60 or value > 180 for value in maximum_values)):
        warnings.warn(
            "Warning: correction ranges are typically between 60 and"
            + " 180 mg/dL; continuing anyway")

    if any(
            start > end for (start, end) in
            list(
                zip(
                    start_times, end_times
                    )
                )[:-1]):
        warnings.warn(
            "Error: correction range start times cannot be greater than range"
            + " end times; stopping run"
        )
        return False

    return True



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jun 21 09:57:26 2019

@author: annaquinlan

Github URL: https://github.com/tidepool-org/LoopKit/blob/
57a9f2ba65ae3765ef7baafe66b883e654e08391/LoopKit/InsulinKit/DoseEntry.swift
"""
from pyloopkit.date import time_interval_since
from pyloopkit.dose import DoseType


def net_basal_units(type_, value, start, end, scheduled_basal_rate, delivered_units):
    """ Find the units of insulin delivered, net of any scheduled basal rate
        (if dose is a temp basal)

    Arguments:
    type_ -- type of dose (basal, bolus, suspend, etc)
    value -- if bolus: amount given, if temp basal: temp rate (U/hr)
    start -- datetime object representing start of dose
    end -- datetime object representing end of dose
    scheduled_basal_rate -- the rate scheduled during the time the dose was
                            given (0 for boluses)
    delivered_units -- units actually delivered by pump

    Output:
    Bolus amount (if a bolus), or basal units given, net of whatever the
    scheduled basal is
    """
    MINIMUM_MINIMED_INCREMENT = 20

    if type_ == DoseType.bolus:
        return delivered_units if delivered_units is not None else value

    elif type_ == DoseType.basal:
        return 0

    hours_ = hours(end, start)

    if hours_ < 0:
        return 0

    if type_ == DoseType.suspend:
        scheduled_units = -scheduled_basal_rate * hours_
    else:
        scheduled_units = (value - scheduled_basal_rate) * hours_

    # round to the basal increments that the pump supports
    return delivered_units if delivered_units is not None else round(scheduled_units * MINIMUM_MINIMED_INCREMENT) / MINIMUM_MINIMED_INCREMENT


def total_units_given(type_, value, start, end):
    """ Find total units given for a dose """
    if type_ in [DoseType.bolus, DoseType.suspend]:
        return value

    return value * hours(end, start)


def hours(start_date, end_date):
    """ Find hours between two dates for the purposes of calculating basal
        delivery
    """
    return abs(time_interval_since(end_date, start_date))/3600  # secs -> hrs



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jun 14 09:21:47 2019

@author: annaquinlan

Github URL: https://github.com/tidepool-org/LoopKit/blob/
57a9f2ba65ae3765ef7baafe66b883e654e08391/LoopKit/GlucoseEffect.swift
"""


class GlucoseEffect:
    """ Constructs a glucose effect

    Attributes:
    start_date -- date and time of the effect
    quantity -- glucose value (mg/dL)
    """
    def __init__(self, start_date, quantity):
        self.start_date = start_date
        self.quantity = quantity

    def __lt__(self, other):
        return self.start_date < other.start_date
